/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include "config.h"

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <glib.h>

#include <stdlib.h> /* atoi */
#include "util.h"
#include "conf.h"
#include "conf_xml.h"
#include "lj_xml.h"

/* XXX gross hack.  why does the normal xmlFree cause segfaults? */
#ifdef G_OS_WIN32
#define xmlFree(x) g_free(x)
#endif

#define xmlGetString lj_xmlGetString
#define xmlAddTN(node, name, value) xmlNewTextChild(node, NULL, name, value)

/*static xmlNodePtr
ljxmlAddTxt(xmlNodePtr node, char *name, char *val) {
	if (val)
		return xmlNewTextChild(node, NULL, name, val);
	return NULL;
}*/

static GList*
parsedirlist(const char *base, void* (*fn)(const char*)) {
	GList *list = NULL;
	GDir *dir;
	const char *dirname;
	char *path;
	void *data;

	dir = g_dir_open(base, 0, NULL);
	if (!dir)
		return NULL;

	dirname = g_dir_read_name(dir);
	while (dirname) {
		path = g_build_filename(base, dirname, NULL);
		data = fn(path);
		if (data)
			list = g_list_append(list, data);
		g_free(path);
		dirname = g_dir_read_name(dir);
	}
	g_dir_close(dir);

	return list;
}

static void*
parsedirxml(const char *dirname, void* (*fn)(xmlDocPtr, xmlNodePtr, void*), void *data) {
	xmlDocPtr doc;
	xmlNodePtr node;
	void *ret;
	char *path;

	path = g_build_filename(dirname, "conf.xml", NULL);

	if (!g_file_test(path, G_FILE_TEST_EXISTS)) {
		g_print("%s doesn't exist\n", path);
		g_free(path);
		return NULL;
	}

	doc = xmlParseFile(path);
	g_free(path);
	if (doc == NULL) {
		fprintf(stderr, _("error parsing configuration file.\n"));
		return NULL;
	}

	/* we get the root element instead of the doc's first child
	 * because XML allows comments outside of the root element. */
	node = xmlDocGetRootElement(doc);
	if (node == NULL) {
		fprintf(stderr, _("empty document.\n"));
		xmlFreeDoc(doc);
		return NULL;
	}

	ret = fn(doc, node, data);
	xmlFreeDoc(doc);

	return ret;
}


/* we use a bunch of macro magic to make this simpler.
 * (see xml_macros.h) 
 */

/* this should match the enum in conf.h */
static char* geometry_names[] = {
	"main",
	"login",
	"friends",
	"friendgroups",
	"console",
	"manager",
	"cffloat",
	"offline",
	"preview"
};

static void
parsegeometry(Geometry *geom, xmlDocPtr doc, xmlNodePtr node) {
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		XML_GET_INT("x", geom->x)
		XML_GET_INT("y", geom->y)
		XML_GET_INT("width", geom->width)
		XML_GET_INT("height", geom->height)
		XML_GET_INT("panedpos", geom->panedpos)
		XML_GET_END("parsegeometry")
	}
}

static void
parsegeometries(Configuration *c, xmlDocPtr doc, xmlNodePtr node) {
	char *window;
	int i;
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		if (xmlStrcmp(node->name, "geometry") == 0) {
			window = xmlGetProp(node, "window");
			if (!window) continue;

			for (i = 0; i < GEOM_COUNT; i++) {
				if (g_ascii_strcasecmp(window, geometry_names[i]) == 0) {
					parsegeometry(&c->geometries[i], doc, node);
					break;
				}
			}
			xmlFree(window);
		}
	}
}

static GList*
parsepickws(xmlDocPtr doc, xmlNodePtr node) {
	GList *pickws = NULL;
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		XML_GET_LIST("pickw", pickws, xmlGetString)
		XML_GET_END("parsepickws")
	}
	return pickws;
}

static GList*
parseusejournals(xmlDocPtr doc, xmlNodePtr node) {
	GList *l = NULL;
	char *journal;
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		if (xmlStrcmp(node->name, "usejournal") == 0) {
			journal = xmlGetString(doc, node);
			l = g_list_append(l, journal);
		} else
		XML_GET_END("usejournals")
	}
	return l;
}

static GList*
parsefriendgroups(xmlDocPtr doc, xmlNodePtr node) {
	GList *fgs = NULL;
	int id;
	FriendGroup *fg;
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		if (xmlStrcmp(node->name, "friendgroup") == 0) {
			if (lj_xmlGetIntProp(node, "id", &id)) {
				fg = g_new0(FriendGroup, 1);
				
				fg->id = id;
				fg->name = lj_xmlGetString(doc, node);
				lj_xmlGetIntProp(node, "ispublic", &fg->ispublic);
				fgs = g_list_append(fgs, fg);
			}
		} else
		XML_GET_END("parsefriendgroups")
	}
	return fgs;
}

static User*
parseuser(xmlDocPtr doc, xmlNodePtr node) {
	User *user;
	user = g_new0(User, 1);
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		XML_GET_STR("username", user->username)
		XML_GET_STR("fullname", user->fullname)
		XML_GET_STR("password", user->password)
		XML_GET_FUNC("pickws", user->pickws, parsepickws)
		XML_GET_FUNC("usejournals", user->usejournals, parseusejournals)
		XML_GET_FUNC("friendgroups", user->friendgroups, parsefriendgroups)
		XML_GET_UINT32("cfmask", user->cfmask)
		XML_GET_END("parseuser")
	}
	user->remember_user = TRUE;
	if (user->password)
		user->remember_password = TRUE;
	return user;
}

static void*
parseuserdir(const char *dirname) {
	return parsedirxml(dirname, (void* (*)(xmlDocPtr, xmlNodePtr, void*))parseuser, NULL);
}
static GList*
parseusersdir(char *base) {
	return parsedirlist(base, parseuserdir);
}

static GList*
parsemoods(xmlDocPtr doc, xmlNodePtr node) {
	GList *moods = NULL;
	NameIDHash *nid;

	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		if (xmlStrcmp(node->name, "mood") == 0) {
			char *idstr, *name;
			idstr = xmlGetProp(node, "id");
			name = xmlGetString(doc, node);

			/* an older bug causes some bad moods to be stored.
			 * filter them out. */
			if (idstr == NULL || name == NULL) {
				if (idstr) xmlFree(idstr);
				if (name) xmlFree(name);
				continue;
			}
			nid = g_new0(NameIDHash, 1);
			nid->id = atoi(idstr);
			xmlFree(idstr);
			nid->name = name;

			moods = g_list_append(moods, nid);
		} else XML_GET_END("parsemoods")
	}
	return moods;
}

static void*
parseserver(xmlDocPtr doc, xmlNodePtr node, void *data) {
	Server *server;
	char *username = NULL;
	char *userspath;

	server = server_new();

	userspath = g_build_filename(data, "users", NULL);
	server->users = parseusersdir(userspath);
	g_free(userspath);

	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		XML_GET_STR("name", server->name)
		XML_GET_STR("url", server->url)
		//XML_GET_FUNC("users", server->users, parseusers)
		XML_GET_INT("protocolversion", server->protocolversion)
		XML_GET_STR("currentuser", username)
		XML_GET_FUNC("moods", server->moods, parsemoods)
		XML_GET_END("parseserver")
	}

	/* after we've parsed all of the users, scan for the current one. */
	if (username) {
		server->usercur = server_get_user_by_username(server, username);
		g_free(username);
	}

	return server;
}

static void*
parseserverdir(const char *dirname) {
	return parsedirxml(dirname, parseserver, (void*)dirname);
}
static GList*
parseserversdir(char *base) {
	return parsedirlist(base, parseserverdir);
}

static void
parseoptions(Configuration *c, xmlDocPtr doc, xmlNodePtr node) {
	Options *options = &c->options;
#define READOPTION(x) XML_GET_BOOL(#x, options->x)
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		XML_GET_BOOL("revertusejournal", options->revertusejournal)
		XML_GET_BOOL("autosave", options->autosave)
		XML_GET_BOOL("netdump", options->netdump)
		XML_GET_BOOL("nofork", options->nofork)
		XML_GET_BOOL("useproxy", options->useproxy)
		XML_GET_BOOL("useproxyauth", options->useproxyauth)
#ifdef HAVE_GTKSPELL
		XML_GET_BOOL("usespellcheck", options->usespellcheck)
#endif
		XML_GET_BOOL("cfautostart", options->cfautostart)
		XML_GET_BOOL("cfusemask", options->cfusemask)
		XML_GET_BOOL("cfdocklet", options->cfdocklet)
		XML_GET_BOOL("cffloat", options->cffloat)
		XML_GET_BOOL("cffloatraise", options->cffloatraise)
		XML_GET_BOOL("friends_statsvis", options->friends_statsvis)
		XML_GET_BOOL("friends_filtervis", options->friends_filtervis)
		XML_GET_BOOL("allowmultipleinstances", options->allowmultipleinstances)
		READOPTION(smartquotes)

		/* XXX backward compat */
		XML_GET_BOOL("cfautofloat", options->cffloat)
		XML_GET_BOOL("cfautofloatraise", options->cffloatraise)
		XML_GET_END("parseoptions")
	}
}

static void
parsesecurity(Configuration *c, xmlDocPtr doc, xmlNodePtr node) {
	char *typestr;
	typestr = xmlGetProp(node, "type");
	if (!typestr) return;
	security_load_from_strings(&c->defaultsecurity, typestr, NULL);
	xmlFree(typestr);
}

static GList*
parsequietdlgs(xmlDocPtr doc, xmlNodePtr node) {
	GList *quiet_dlgs = NULL;
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		XML_GET_LIST("quiet_dlg", quiet_dlgs, xmlGetString)
		XML_GET_END("parsequietdlgs")
	}
	return quiet_dlgs;
}

#ifndef G_OS_WIN32
static void
parseproxyauth(Configuration *c, xmlDocPtr doc, xmlNodePtr node) {
	for (node = node->xmlChildrenNode; node != NULL; node = node->next) {
		XML_GET_STR("username", c->proxyuser)
		XML_GET_STR("password", c->proxypass)
		XML_GET_END("parseproxyauth")
	}
}

static void
parsemusicsource(Configuration *c, xmlDocPtr doc, xmlNodePtr node) {
	char *type;
	
	type = xmlGetProp(node, "type");
	if (!type) return;

	if (g_ascii_strcasecmp(type, "xmms") == 0) {
		c->music_source = MUSIC_SOURCE_XMMS;
	} else if (g_ascii_strcasecmp(type, "command") == 0) {
		c->music_source = MUSIC_SOURCE_COMMAND;
		c->music_command = xmlGetString(doc, node);
	}
	xmlFree(type);
}
#endif

static void*
parseconf(xmlDocPtr doc, xmlNodePtr node, void *data) {
	char *servername = NULL;
	char *serverspath;
	Configuration *c = data;

#define XML_GET_CONF(key, func) XML_GET_IF(key, func(c, doc, node);) 
	node = node->xmlChildrenNode;
	for (; node != NULL; node = node->next) {
		XML_GET_STR("currentserver", servername)
		XML_GET_CONF("geometries", parsegeometries)
		XML_GET_CONF("options", parseoptions)
		XML_GET_CONF("defaultsecurity", parsesecurity)
		XML_GET_STR("uifont", c->uifont)
#ifdef HAVE_GTKSPELL
		XML_GET_STR("spell_language", c->spell_language)
#endif /* HAVE_GTKSPELL */
#ifndef G_OS_WIN32
		XML_GET_STR("proxy", c->proxy)
		XML_GET_CONF("proxyauth", parseproxyauth)
		XML_GET_STR("spawncommand", c->spawn_command)
		XML_GET_CONF("musicsource", parsemusicsource)
#endif /* G_OS_WIN32 */
		XML_GET_INT("cfuserinterval", c->cfuserinterval)
		XML_GET_INT("cfthreshold", c->cfthreshold)
		XML_GET_FUNC("quiet_dlgs", app.quiet_dlgs, parsequietdlgs)
		XML_GET_END("conf_xml_read")
	}

#ifdef HAVE_GTKSPELL
	if (!c->spell_language || strlen(c->spell_language) < 2) {
		g_free(c->spell_language);
		c->spell_language = g_strdup("en");
	}
#endif /* HAVE_GTKSPELL */
	serverspath = g_build_filename(app.conf_dir, "servers", NULL);
	c->servers = parseserversdir(serverspath);
	g_free(serverspath);

	if (servername) {
		c->servercur = conf_server_by_name(c, servername);
		g_free(servername);
	}
	return NULL;
}

int
conf_read(Configuration *c, char *path) {
	parsedirxml(path, parseconf, c);
	return 0;
}

static gboolean
writeuser(Server *server, User *user) {
	GList *l;
	xmlDocPtr doc;
	xmlNodePtr xmlList, child;
	FriendGroup *fg;
	char buf[10];
	char *path;
	xmlNodePtr node;

	if (!user->remember_user)
		return FALSE;

	path = g_build_filename(app.conf_dir, "servers", server->name,
			"users", user->username, "conf.xml", NULL);
	if (!verify_path(path, FALSE, NULL))
		return FALSE;

	lj_xmlNewDoc(&doc, &node, "user");

	xmlAddTN(node, "username", user->username);
	if (user->remember_password)
		xmlAddTN(node, "password", user->password);
	if (user->fullname)
		xmlAddTN(node, "fullname", user->fullname);
	if (user->cfmask) {
		char buf[10];
		g_snprintf(buf, 10, "%d", user->cfmask);
		xmlAddTN(node, "cfmask", buf);
	}


	if (user->pickws) {
		xmlList = xmlNewChild(node, NULL, "pickws", NULL);
		for (l = user->pickws; l != NULL; l = l->next) {
			xmlAddTN(xmlList, "pickw", (char*)l->data);
		}
	}

	if (user->usejournals) {
		xmlList = xmlNewChild(node, NULL, "usejournals", NULL);
		for (l = user->usejournals; l != NULL; l = l->next) {
			xmlAddTN(xmlList, "usejournal", l->data);
		}
	}

	if (user->friendgroups) {
		xmlList = xmlNewChild(node, NULL, "friendgroups", NULL);
		for (l = user->friendgroups; l != NULL; l = l->next) {
			fg = (FriendGroup*)l->data;
			child = xmlAddTN(xmlList, "friendgroup", fg->name);
			g_snprintf(buf, 10, "%d", fg->id);
			xmlSetProp(child, "id", buf);
			g_snprintf(buf, 10, "%d", fg->ispublic);
			xmlSetProp(child, "ispublic", buf);
		}
	}

	if (xmlSaveFormatFile(path, doc, TRUE) < 0) 
		fprintf(stderr, "Error saving to %s!?\n", path);
	xmlFreeDoc(doc);
	g_free(path);

	return TRUE;
}

static gboolean
writeserverdir(Server *server) {
	xmlDocPtr doc;
	xmlNodePtr node;
	GList *l;
	char *path;
	xmlNodePtr servernode;

	path = g_build_filename(app.conf_dir, "servers", server->name, "conf.xml", NULL);
	if (!verify_path(path, FALSE, NULL))
		return FALSE;

	lj_xmlNewDoc(&doc, &servernode, "server");

	xmlAddTN(servernode, "name", server->name);
	xmlAddTN(servernode, "url", server->url);
	lj_xmlAddInt(servernode, "protocolversion", server->protocolversion);

	for (l = server->users; l != NULL; l = l->next) 
		writeuser(server, l->data);

	if (server->usercur)
		xmlAddTN(servernode, "currentuser", server->usercur->username);

	node = xmlNewChild(servernode, NULL, "moods", NULL);
	for (l = server->moods; l != NULL; l = l->next) {
		Mood *mood = (Mood*)l->data;
		char buf[10];
		xmlNodePtr moodnode = xmlAddTN(node, "mood", mood->name);
		
		g_snprintf(buf, 10, "%d", mood->id);
		xmlSetProp(moodnode, "id", buf);
	}

	if (xmlSaveFormatFile(path, doc, TRUE) < 0) 
		fprintf(stderr, "Error saving to %s!?\n", path);
	xmlFreeDoc(doc);
	
	return TRUE;
}

static void
writegeometry(Geometry *geometry, char *name, xmlNodePtr node) {
	node = xmlNewChild(node, NULL, "geometry", NULL);
	xmlSetProp(node, "window", name);
	lj_xmlAddInt(node, "x", geometry->x);
	lj_xmlAddInt(node, "y", geometry->y);
	lj_xmlAddInt(node, "width",  geometry->width);
	lj_xmlAddInt(node, "height", geometry->height);
	if (geometry->panedpos > 0)
		lj_xmlAddInt(node, "panedpos", geometry->panedpos);
}

static void
writeoptions(Options *options, xmlNodePtr node) {
#define WRITEOPTION(x) if (options->x) xmlNewChild(node, NULL, #x, NULL);
	WRITEOPTION(revertusejournal);
	WRITEOPTION(autosave);
	WRITEOPTION(netdump);
	WRITEOPTION(nofork);
	WRITEOPTION(useproxy);
	WRITEOPTION(useproxyauth);
#ifdef HAVE_GTKSPELL
	WRITEOPTION(usespellcheck);
#endif
	WRITEOPTION(cfautostart);
	WRITEOPTION(cfusemask);
	WRITEOPTION(cfdocklet);
	WRITEOPTION(cffloatraise);
	WRITEOPTION(cffloat);
	WRITEOPTION(friends_statsvis);
	WRITEOPTION(friends_filtervis);
	WRITEOPTION(allowmultipleinstances);
	WRITEOPTION(smartquotes);
}

int
conf_write(Configuration *c, char *base) {
	xmlDocPtr doc;
	xmlNodePtr root, node;
	GList *l;
	int i;
	char *path;

	lj_xmlNewDoc(&doc, &root, "configuration");
	xmlSetProp(root, "version", "1");

	for (l = c->servers; l != NULL; l = l->next) {
		writeserverdir(l->data);
	}

	if (c->servercur)
		xmlAddTN(root, "currentserver", c->servercur->name);

	node = xmlNewChild(root, NULL, "geometries", NULL);
	for (i = 0; i < GEOM_COUNT; i++) {
		if (c->geometries[i].width > 0) {
			writegeometry(&c->geometries[i], geometry_names[i], node);
		}
	}

	node = xmlNewChild(root, NULL, "options", NULL);
	writeoptions(&c->options, node);

	if (c->uifont)
		xmlAddTN(root, "uifont", c->uifont);

#ifdef HAVE_GTKSPELL
	if (c->spell_language)
		xmlAddTN(root, "spell_language", c->spell_language);
#endif
#ifndef G_OS_WIN32
	if (c->proxy)
		xmlAddTN(root, "proxy", c->proxy);
	if (c->proxyuser) {
		node = xmlNewChild(root, NULL, "proxyauth", NULL);
		xmlAddTN(node, "username", c->proxyuser);
		xmlAddTN(node, "password", c->proxypass);
	}

	if (c->spawn_command)
		xmlAddTN(root, "spawncommand", c->spawn_command);

	if (c->music_source == MUSIC_SOURCE_COMMAND) {
		node = xmlAddTN(root, "musicsource", c->music_command);
		xmlSetProp(node, "type", "command");
	}
#endif

	if (c->cfuserinterval) {
		char buf[20];
		g_snprintf(buf, 20, "%d", c->cfuserinterval);
		xmlAddTN(root, "cfuserinterval", buf);
	}

	if (c->cfthreshold) {
		char buf[20];
		g_snprintf(buf, 20, "%d", c->cfthreshold);
		xmlAddTN(root, "cfthreshold", buf);
	}

	if (app.quiet_dlgs) {
		xmlNodePtr xmlList = xmlNewChild(root, NULL, "quiet_dlgs", NULL);
		for (l = app.quiet_dlgs; l != NULL; l = l->next) {
			xmlAddTN(xmlList, "quiet_dlg", (char*)l->data);
		}
	}

	if (c->defaultsecurity.type != SECURITY_PUBLIC) {
		char *type = NULL;
		if (c->defaultsecurity.type == SECURITY_CUSTOM)
			c->defaultsecurity.type = SECURITY_PRIVATE;
		security_save_to_strings(&c->defaultsecurity, &type, NULL);
		if (type) {
			node = xmlNewChild(root, NULL, "defaultsecurity", NULL);
			xmlSetProp(node, "type", type);
			g_free(type);
		}
	}

	path = g_build_filename(base, "conf.xml", NULL);
	if (xmlSaveFormatFile(path, doc, TRUE) < 0) 
		fprintf(stderr, "Error saving to %s!?\n", path);
	xmlFreeDoc(doc);
	g_free(path);
	return 0;
}


