/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include "config.h"

#include <gtk/gtk.h>
#include <glib.h>

#define _GNU_SOURCE
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

#ifndef G_OS_WIN32
#include <unistd.h>
#endif

#include <stdlib.h>  /* for exit (and __argc and __argv on windows) */

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <locale.h>

#include "network.h"
#include "conf.h"
#include "conf_xml.h"
#include "util.h"
#include "login.h"
#include "lj.h"
#include "spawn.h"
#include "icons.h"
#include "checkfriends.h"
#include "remote.h"

enum {
	LJ_ARGS_OPTION,
	LJ_ARGS_USER_UNKNOWN,
	LJ_ARGS_NO_USER,
	LJ_ARGS_NO_POST,
	LJ_ARGS_CF_OPTION,

	LJ_ARGS_FAILED
};

#define LJ_ARGS lj_args_error_quark()

static GQuark
lj_args_error_quark (void) {
	static GQuark q = 0;
	if (q == 0)
		q = g_quark_from_static_string ("lj-args-error-quark");

	/* silence a compiler warning. */
	if (0)
		lj_args_error_quark();
	return q;
}

static void
print_header(char *name) {
	g_print(_("LogJam %s\nCopyright (C) 2000-2003 Evan Martin\n\n"), PACKAGE_VERSION);
}

static void
print_help(char *name) {
#include "helptext.h" /* autogenerated */
	print_header(name);
	g_print(_(help), name);
#ifndef HAVE_GETOPT_LONG
	g_print(_("(Long options aren't supported on this system.)\n"));
#endif
}

static int
post_cli(Entry *entry) {
	NetRequest *request;
	NetResult *result;

	time_t curtime_time_t = time(NULL);
	struct tm *curtime = localtime(&curtime_time_t);

	/* silence a compiler warning. */
	if (0)
		post_cli(entry);

	entry->time = *curtime;

	request = net_request_new("postevent");
	entry_set_request_fields(entry, request);
	result = net_request_run_cli(request);
	net_request_free(request);
	
	if (net_result_succeeded(result)) {
		g_print(_("Success.\n"));
	} else {
		net_result_free(result);
		return -1;
	}
	net_result_free(result);

	return 0;
}

void set_defaults() {
#ifndef G_OS_WIN32
	if (conf.spawn_command == NULL)
		conf.spawn_command = g_strdup(spawn_commands[0]);
#endif

	conf_verify_a_server_exists();

	if (conf.servercur == NULL) {
		conf.servercur = conf.servers->data;
	}

	if (conf.cfuserinterval == 0)
		conf.cfuserinterval = 30;
	cf_threshold_normalize(&conf.cfthreshold);

#ifndef G_OS_WIN32
#if XMMS_TYPE == XMMS_TYPE_NONE
	conf.music_source = MUSIC_SOURCE_COMMAND;
#endif
#endif
}

#ifdef HAVE_GETOPT_H
/* may return false with no error, eg when -h is issued on the
 * command line. check err to see if an error occured. */
static gboolean
parse_options(int argc, char* argv[], LogJamDoc *doc, GError **err) {
#include "cmdlineopts_short.h"
#ifdef HAVE_GETOPT_LONG
#include "cmdlineopts_long.h"
#endif
	int c = 0;
	char *username = NULL, *password = NULL, *postas = NULL, *cf_mode = NULL;
	gchar *opt_subject = NULL, *opt_pickeyword = NULL;
	gboolean postcli = FALSE;
	gboolean cf_cli = FALSE;
	gboolean use_editor = FALSE;
	GError *tmp_error = NULL;

	while (c != -1) {

#ifdef HAVE_GETOPT_LONG
		c = getopt_long(argc, argv, short_options, long_options, NULL);
#else
		c = getopt(argc, argv, short_options);
#endif
		
		switch (c) {
			case 'h':
				print_help(argv[0]);
				return FALSE;
			case 'v':
				print_header(argv[0]);
				return FALSE;
			case 'c':
				postcli = TRUE;
				break;
			case 'u':
				username = optarg;
				break;
			case 'a':
				postas = optarg;
				break;
			case 'p':
				password = optarg;
				break;
			case 'n':
				app.skiplogin = TRUE;
				break;
			case 's':
				opt_subject          = optarg;
				break;
			case 'k':
				opt_pickeyword       = optarg;
				break;
			case 'e':
				use_editor = TRUE;
				break;
			case 'r':
				cf_cli  = TRUE;
				cf_mode = optarg;
				break;
			case 'q':
				app.quiet = TRUE;
				break;
			case '?': /* unknown option character. */
			case ':': /* missing option character. */
				g_set_error(err, LJ_ARGS, LJ_ARGS_OPTION,
						_("Unknown or missing option character"));
				return FALSE;
		}
	}

	if (optind < argc) { /* read entry from file (possibly override meta) */
		if (strcmp(argv[optind], "-") == 0) {
			Entry *entry;
			entry = entry_new_from_stdin(err);
			if (!entry)
				return FALSE;
			logjam_doc_load_entry(doc, entry);
			entry_free(entry);
		} else {
			if (!logjam_doc_load_file(doc, argv[optind],
					ENTRY_FILE_AUTODETECT, err))
				return FALSE;
		}
	}

	if (username) {
		User *u;

		u = server_get_user_by_username(conf_cur_server(), username);
		if (!u) {
			g_set_error(err, LJ_ARGS, LJ_ARGS_USER_UNKNOWN,
					_("Unknown user: %s.\n"), username);
			return FALSE;
		}

		if (password) {
			string_replace(&u->password, g_strdup(password));
		} 
		if (!u->password && (postcli || !app.skiplogin)) {
			if ((password = getpass(_("Password: "))) != NULL)
				u->password = g_strdup(password);
		}
		conf_cur_user() = u;
		app.autolog = TRUE;
	}

	if (postas) 
		string_replace(&conf.usejournal, g_strdup(postas));

#if 0
	XXX doc

	if (use_editor && !entry_edit_with_usereditor(entry, err)) {
		return FALSE;
	}

	if (postcli) {
		if (!conf_cur_user()) {
			g_set_error(err, LJ_ARGS, LJ_ARGS_NO_USER, _("No user specified"));
			return FALSE;
		}
		if (entry->event == NULL) {
			g_set_error(err, LJ_ARGS, LJ_ARGS_NO_POST, _("No post specified"));
			g_print(_("No post specified!\n"));
			return FALSE;
		}

		if (post_cli(entry) == 0) /* if the post succeeded, we're done. */
			return FALSE;
	}
#endif

	if (cf_cli) {
		gchar *ljid;
		
		if (!conf_cur_user()) {
			g_set_error(err, LJ_ARGS, LJ_ARGS_NO_USER, _("No user specified"));
			return FALSE;
		}

		ljid = g_strdup_printf("%s@%s",
				conf_cur_user()->username, conf_cur_server()->name);
		app.cli = TRUE;

		if (cf_mode == NULL) {
			if (checkfriends_cli(ljid)) {
				g_print("1\n"); exit(0); /* NEW */
			} else {
				g_print("0\n"); exit(1); /* no new entries, or some error */
			}
		}
		if (!g_ascii_strncasecmp("purge", cf_mode,
				strlen("purge"))) {
			checkfriends_cli_purge(ljid);
			g_print(_("Checkfriends information for %s purged.\n"), ljid);
			exit(0);
		}
		g_set_error(err, LJ_ARGS, LJ_ARGS_CF_OPTION,
				_("Unknown argument %s to --checkfriends\n"), cf_mode);
		return FALSE;
	}

	/* send the remote command.
	 * you'd think we'd want to conditionally do this on the configuration
	 * option, but you'd be wrong.  :)
	 * if the setting in the conf is "allow multiple instances", but then
	 * the user changes the setting to "only one instance" in the ui and then
	 * runs another copy of logjam, it's counterintuitive for it to ignore that
	 * changed ui setting, even if it hadn't got propogated to the
	 * configuration file yet.
	 * so we always try to use the socket.  the conf setting is only
	 * used when we consider *creating* the socket. */
	if (remote_send_user(username, app.skiplogin, &tmp_error)) {
		/* we popped up another window, so we're done. */
		g_print(_("Bringing already-running LogJam instance to the "
				  "foreground...\n"));
		return FALSE;
	} else if (tmp_error) {
		g_propagate_error(err, tmp_error);
		return FALSE;
	}

	return TRUE;
}
#endif /* HAVE_GETOPT_H */

static void
init_app(int *argc, gchar *argv[]) {
	gint i, shift = 0;
	
	memset(&app, 0, sizeof(Application));
	app.programname      = argv[0]; /* not the same as PROGRAMNAME */
	app.autosave_ok      = TRUE;
	app.autoloaddraft_ok = TRUE;

	app.remote = logjam_remote_new();
	
	/* treatment of configuration dir; must be done before read_conf */
	/* XXX: I wish there were a more concise way to do this (Perl!:) */
	for (i = 1; i < *argc; i++) {                    /* look for arg */
		if (!strcmp(argv[i], "--conf-dir")) {
			app.conf_dir = argv[i+1];
			shift = 2;
			break;
		}
		if (g_strrstr(argv[i], "--conf-dir=")) {
			app.conf_dir = argv[i] + strlen("--conf-dir=");
			shift = 1;
			break;
		}
	}
	if (shift) {                           /* shift what comes after */
		*argc -= shift;
		for (; i < *argc; i++) {
			argv[i] = argv[i + shift];
		}
	}

	if (!app.conf_dir) {
		app.conf_dir = g_strdup_printf("%s/.logjam", g_get_home_dir());
	}
}

int
main(int argc, char* argv[]) {
	gboolean has_gtk;
	LogJamDoc *doc;
	gchar *accelpath;
	GError *err;
	err = NULL;

	/* This call actually initializes the locale settings,
	 * and doesn't just clear LC_ALL (which is sorta what
	 * it looks like it's doing). */
	setlocale(LC_ALL, "");

#ifdef ENABLE_NLS
	bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
	/* GTK always wants UTF-8 strings.
	 * gettext will try to translate to the current locale's codeset
	 * unless we tell it to leave it in UTF-8. */
	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
	textdomain(GETTEXT_PACKAGE);
#endif

#ifdef G_OS_WIN32
	/* calls to these thread inits must happen before any other
	 * g* calls. */
	if (!g_thread_supported()) g_thread_init(NULL);
	gdk_threads_init();
	gdk_threads_enter();

	gtk_rc_add_default_file("gtkrc");
#endif
	has_gtk = gtk_init_check(&argc, &argv);
	g_type_init();

	init_app(&argc, argv); /* should be called before conf_read */
	conf_read(&conf, app.conf_dir);
	set_defaults();

	doc = logjam_doc_new();
#ifdef HAVE_GETOPT_H
	if (!parse_options(argc, argv, doc, &err)) {
		if (err) {
			g_print(_("Error: %s\n"), err->message);
			g_error_free(err);
			return -1;
		}
		return 0;
	}
#endif

	if (!has_gtk) {
		g_print(_("GTK init failed -- are you in X?\n\n"));
		print_help(argv[0]);
		return -1;
	}

	accelpath = g_build_filename(app.conf_dir, "accels", NULL);
	gtk_accel_map_load(accelpath);

	icons_initialize();
	lj_run(doc);

	g_object_unref(G_OBJECT(app.remote));

	conf_write(&conf, app.conf_dir);
	gtk_accel_map_save(accelpath);

	g_free(accelpath);

#ifdef G_OS_WIN32
	gdk_threads_leave();
#endif

	return 0;
}

#ifdef G_OS_WIN32
int _stdcall WinMain(void *hInst, void *hPrev, char* lpCmdLine, int nCmdShow) {
	return main(__argc, __argv);
}
#endif

