/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include <config.h>
#include <gtk/gtk.h>
#ifdef HAVE_GTKSPELL
#include <gtkspell/gtkspell.h>
#endif /* HAVE_GTKSPELL */

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#ifndef G_OS_WIN32
#include <unistd.h> /* unlink */
#else
#include <io.h>
#endif /* G_OS_WIN32 */

#include "network.h"
#include "conf.h"
#include "util.h"
#include "livejournal.h"
#include "login.h"
#include "logjam-doc.h"
#include "lj.h"
#include "menu.h"
#include "security.h"
#include "icons.h"
#include "meta.h"
#include "checkfriends.h"
#include "usejournal.h"
#include "draftstore.h"
#include "remote.h"

#undef USE_STRUCTUREDTEXT

static void lj_update_title(LJWin *ljw);
static void lj_autosave_delete();
static void lj_new_doc(LJWin *ljw, LogJamDoc *doc);
static void lj_update_actions(LJWin *ljw);

static const int LJ_AUTOSAVE_INTERVAL = 5*1000;

gboolean
lj_confirm_lose_entry(LJWin *ljw) {
	GtkWidget *dlg;
	int ret;

	if (!logjam_doc_get_dirty(ljw->doc))
		return TRUE;

	dlg = gtk_message_dialog_new(GTK_WINDOW(ljw),
			GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION,
			GTK_BUTTONS_NONE,
			_("Your entry has unsaved changes.  Save them now?"));
	gtk_dialog_add_buttons(GTK_DIALOG(dlg),
			_("Don't Save"), GTK_RESPONSE_NO,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_SAVE, GTK_RESPONSE_YES,
			NULL);
	gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_YES);
ask:
	ret = gtk_dialog_run(GTK_DIALOG(dlg));

	switch (ret) {
		case GTK_RESPONSE_NO: /* don't save */
			ret = TRUE;
			break;
		case GTK_RESPONSE_YES: /* save */
			if (!lj_save(ljw))
				goto ask;
			ret = TRUE;
			break;
		case GTK_RESPONSE_CANCEL: /* cancel */
		default:
			ret = FALSE;
			break;
	}
	gtk_widget_destroy(dlg);
	return ret;
}

gboolean
lj_save_as_file(LJWin *ljw) {
	GtkWidget *filesel;
	GError *err = NULL;
	
	filesel = lj_filesel_new(_("Save Entry As"), GTK_WINDOW(ljw));

	if (gtk_dialog_run(GTK_DIALOG(filesel)) == GTK_RESPONSE_OK) {
		const gchar *filename;
		filename = gtk_file_selection_get_filename(
				GTK_FILE_SELECTION(filesel));
		if (!logjam_doc_save_as_file(ljw->doc, filename, &err)) {
			lj_warning(GTK_WINDOW(filesel), _("Unable to save to '%s': %s."),
					filename, err->message);
			g_error_free(err);
			gtk_widget_destroy(filesel);
			return FALSE;
		} else {
			gtk_widget_destroy(filesel);
			return TRUE;
		}
	}
	gtk_widget_destroy(filesel);
	return FALSE;
}

static char*
prompt_draftname(GtkWindow *parent, const char *basename) {
	GtkWidget *dlg, *box, *entry;

	dlg = gtk_dialog_new_with_buttons(_("New Draft Title"),
			parent, GTK_DIALOG_MODAL,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_OK, GTK_RESPONSE_OK,
			NULL);
	gtk_window_set_transient_for(GTK_WINDOW(dlg), parent);
	gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_OK);

	entry = gtk_entry_new();
	if (basename)
		gtk_entry_set_text(GTK_ENTRY(entry), basename);
	gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
	box = labelled_box_new(_("_Draft Title:"), entry);

	lj_dialog_set_contents(GTK_DIALOG(dlg), box);
	gtk_widget_grab_focus(entry);

	if (gtk_dialog_run(GTK_DIALOG(dlg)) == GTK_RESPONSE_OK) {
		char *text;
		text = gtk_editable_get_chars(GTK_EDITABLE(entry), 0, -1);
		gtk_widget_destroy(dlg);
		return text;
	}
	gtk_widget_destroy(dlg);
	return NULL;
}

gboolean
lj_save_as_draft(LJWin *ljw) {
	GError *err = NULL;
	char *sugg, *title;
	gboolean ret;

	sugg = logjam_doc_get_draftname(ljw->doc);
	title = prompt_draftname(GTK_WINDOW(ljw), sugg);
	g_free(sugg);
	if (!title)
		return FALSE;

	if (!logjam_doc_save_as_draft(ljw->doc, title, &err)) {
		lj_warning(GTK_WINDOW(ljw), _("Error saving draft: %s."), err->message);
		g_error_free(err);
		ret = FALSE;
	} else {
		ret = TRUE;
	}
	g_free(title);
	lj_update_actions(ljw);
	return ret;
}

gboolean
lj_save(LJWin *ljw) {
	GError *err = NULL;

	if (logjam_doc_has_save_target(ljw->doc)) {
		if (logjam_doc_would_save_over_nonxml(ljw->doc)) {
			if (!lj_confirm(GTK_WINDOW(ljw), _("Saving File"),
							_("Current file was imported from an non-XML file.  "
							"Saving to this file will overwrite the file with XML content.  "
							"Overwrite the file?"))) {
				return FALSE;
			}
		}

		if (!logjam_doc_save(ljw->doc, &err)) {
			lj_warning(GTK_WINDOW(ljw), _("Error saving: %s."), err->message);
			g_error_free(err);
			return FALSE;
		}
	} else {
		return lj_save_as_file(ljw);
	}
	return TRUE;
}


void
lj_clear_entry(LJWin *ljw) {
	g_object_unref(G_OBJECT(ljw->doc));
	lj_new_doc(ljw, logjam_doc_new());
}

static void
lj_update_actions(LJWin *ljw) {
	int flags;

	flags = logjam_doc_get_flags(ljw->doc);

	if (flags & LOGJAM_DOC_CAN_SAVE) {
		gtk_label_set_text(GTK_LABEL(ljw->laction), _("  Save Changes  "));
	} else if (flags & LOGJAM_DOC_CAN_SUBMIT) {
		gtk_label_set_text(GTK_LABEL(ljw->laction), _("  Submit  "));
		gtk_widget_hide(ljw->msaveserver);
	}

	lj_widget_set_visible(ljw->actionbox, conf.loginok);
	lj_widget_set_visible(ljw->bdelete, flags & LOGJAM_DOC_CAN_DELETE);

	lj_widget_set_visible(ljw->msubmitsep,  conf.loginok);
	lj_widget_set_visible(ljw->msubmit,     conf.loginok);
	lj_widget_set_visible(ljw->msaveserver, conf.loginok && (flags & LOGJAM_DOC_CAN_SAVE));
}

void
entry_changed_cb(LogJamDoc *doc, Entry *entry, LJWin *ljw) {
	lj_update_actions(ljw);
	metamgr_from_entry(METAMGR(ljw->metamgr), entry);
	menu_load_entry(ljw, entry);
}

static void
lj_update_title(LJWin *ljw) {
	char *doctitle, *title;
	doctitle = logjam_doc_get_title(ljw->doc);
	title = g_strdup_printf(_("LogJam - %s%s"), doctitle,
			logjam_doc_get_dirty(ljw->doc) ? "*" : "");
	gtk_window_set_title(GTK_WINDOW(ljw), title);
	g_free(title);
	g_free(doctitle);
}

static void
title_changed_cb(LogJamDoc *doc, LJWin *ljw) {
	lj_update_title(ljw);
}
static void
dirty_changed_cb(LogJamDoc *doc, gpointer foo, LJWin *ljw) {
	lj_update_title(ljw);
}

static void
update_entry_cb(LogJamDoc *doc, Entry *entry, LJWin *ljw) {
	GtkTextBuffer *buffer;
	GtkTextIter start, end;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(ljw->eentry));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
	string_replace(&entry->event, 
			gtk_text_buffer_get_text(buffer, &start, &end, TRUE));
	metamgr_to_entry(METAMGR(ljw->metamgr), entry);
}

static void
undo_cb(GtkWidget *menu_item, LJWin *ljw) {
	if(undo_manager_can_undo(ljw->um))
		undo_manager_undo(ljw->um);
}

static void
redo_cb(GtkWidget *menu_item, LJWin *ljw) {
	if(undo_manager_can_redo(ljw->um))
		undo_manager_redo(ljw->um);
}

static void
populate_entry_popup(GtkTextView *view, GtkMenu *menu, LJWin *ljw) {
	GtkWidget *menu_item;

	g_return_if_fail(menu != NULL);
	g_return_if_fail(GTK_IS_MENU_SHELL(menu));
	g_return_if_fail(ljw != NULL);

	menu_item = gtk_separator_menu_item_new();
	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), menu_item);
	gtk_widget_show(menu_item);

	menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_REDO, NULL);
	gtk_widget_set_sensitive(menu_item, undo_manager_can_redo(ljw->um));
	g_signal_connect(G_OBJECT(menu_item), "activate",
			G_CALLBACK(redo_cb), ljw);
	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), menu_item);
	gtk_widget_show(menu_item);
	
	menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_UNDO, NULL);
	gtk_widget_set_sensitive(menu_item, undo_manager_can_undo(ljw->um));
	g_signal_connect(G_OBJECT(menu_item), "activate",
			G_CALLBACK(undo_cb), ljw);
	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), menu_item);
	gtk_widget_show(menu_item);
}

static void
can_undo_cb(UndoManager *um, gboolean can_undo, LJWin *ljw) {
	gtk_widget_set_sensitive(ljw->mundo, can_undo);
}

static void
can_redo_cb(UndoManager *um, gboolean can_redo, LJWin *ljw) {
	gtk_widget_set_sensitive(ljw->mredo, can_redo);
}

static void
lj_new_doc(LJWin *ljw, LogJamDoc *doc) {
	Entry *entry;

	/* XXX: gtkspell gets confused if you detach its document,
	 * so we detach the spell checker completely, then reattach
	 * it at the end. */
#ifdef HAVE_GTKSPELL
	if (conf.options.usespellcheck && ljw->eentry) {
		GtkSpell *spell;
		spell = gtkspell_get_from_text_view(GTK_TEXT_VIEW(ljw->eentry));
		if (spell)
			gtkspell_detach(spell);
	}
#endif /* HAVE_GTKSPELL */

	ljw->doc = doc;
	g_signal_connect(G_OBJECT(ljw->doc), "title-changed",
			G_CALLBACK(title_changed_cb), ljw);
	g_signal_connect(G_OBJECT(ljw->doc), "notify::dirty",
			G_CALLBACK(dirty_changed_cb), ljw);
	g_signal_connect(G_OBJECT(ljw->doc), "entry-changed",
			G_CALLBACK(entry_changed_cb), ljw);
	g_signal_connect(G_OBJECT(ljw->doc), "update-entry",
			G_CALLBACK(update_entry_cb), ljw);
	lj_update_title(ljw);


	/* if the ui is already up, we need to let everything
	 * know the doc has changed. */
	if (ljw->eentry) {
		undo_manager_detach(ljw->um, ljw->eentry);

		gtk_text_view_set_buffer(GTK_TEXT_VIEW(ljw->eentry),
				logjam_doc_get_text_buffer(ljw->doc));
		entry = entry_new();
		entry_changed_cb(ljw->doc, entry, ljw);
		entry_free(entry);

		undo_manager_attach(ljw->um, ljw->eentry);
		g_signal_connect(G_OBJECT(ljw->um), "can-undo",
			G_CALLBACK(can_undo_cb), ljw);
		g_signal_connect(G_OBJECT(ljw->um), "can-redo",
			G_CALLBACK(can_redo_cb), ljw);
		
#ifdef HAVE_GTKSPELL
		if (conf.options.usespellcheck) {
			GError *err = NULL;
			if (gtkspell_new_attach(GTK_TEXT_VIEW(ljw->eentry),
						conf.spell_language, &err) == NULL) {
				lj_warning(GTK_WINDOW(ljw),
						_("GtkSpell error: %s"), err->message);
				g_error_free(err);
			}
		}
#endif /* HAVE_GTKSPELL */
	}
}

static gboolean
lj_action(LJWin *ljw, Entry *entry, const char *verb, const char *title) {
	NetRequest *request;
	NetResult *result;
	gboolean success;

	if (conf_cur_server()->protocolversion < 1) {
		/* scan the entry and make sure they're sending ASCII. */
		unsigned char *p;
		gboolean invalid = FALSE;

		for (p = entry->event; *p; p++) {
			if (*p > 0x7F) {
				invalid = TRUE;
				break;
			}
		}
		if (invalid) {
			lj_warning(GTK_WINDOW(ljw), _("Error: Your post contains Unicode "
						"characters, but you trying to submit them to a server "
						"that doesn't support Unicode."));
			return FALSE;
		}
	}

	request = net_request_new(verb);
	entry_set_request_fields(entry, request);

	result = net_request_run(request, title, GTK_WINDOW(ljw));
	net_request_free(request);

	success = net_result_succeeded(result);
	net_result_free(result);
	return success;
}
		
void
lj_save_entry_server(LJWin *ljw) {
	Entry *entry;
	entry = logjam_doc_get_entry(ljw->doc);
	if (lj_action(ljw, entry, "editevent", _("Saving changes..."))) {
		lj_clear_entry(ljw);
		lj_autosave_delete();
	}
	entry_free(entry);
}

void
lj_submit_entry(LJWin *ljw) {
	Entry *entry;
	entry = logjam_doc_get_entry(ljw->doc);
	entry->itemid = 0;
	if (lj_action(ljw, entry, "postevent", _("Submitting entry..."))) {
		lj_clear_entry(ljw);
		if (conf.options.revertusejournal)
			usejournal_revert(USEJOURNAL(ljw->usejournal));
		lj_autosave_delete();
	}
	entry_free(entry);
}

static void 
action_cb(GtkWidget *w, LJWin *ljw) {
	int flags = logjam_doc_get_flags(ljw->doc);
	if (flags & LOGJAM_DOC_CAN_SAVE) {
		lj_save_entry_server(ljw);
	} else if (flags & LOGJAM_DOC_CAN_SUBMIT) {
		lj_submit_entry(ljw);
	} else {
		g_warning("action callback, but no default action?\n");
	}
}

static void
delete_draft(LJWin *ljw) {
	DraftStore *ds;
	GError *err = NULL;

	ds = draft_store_new_from_current();

	if (!draft_store_remove_entry(ds,
				logjam_doc_get_entry_itemid(ljw->doc), &err)) {
		lj_warning(GTK_WINDOW(ljw), err->message);
		g_error_free(err);
	} else {
		lj_clear_entry(ljw);
	}

	draft_store_free(ds);
}


static void
delete_server_entry(LJWin *ljw) {
	Entry *entry = logjam_doc_get_entry(ljw->doc);

	string_replace(&entry->event, g_strdup(""));
	if (lj_action(ljw, entry, "editevent", _("Deleting entry...")))
		lj_clear_entry(ljw);
	entry_free(entry);
}
	
static void 
delete_cb(GtkWidget *w, LJWin *ljw) {
	gint type = logjam_doc_get_entry_type(ljw->doc); /* defer making a copy */
	const gchar *confirm_text = type == ENTRY_DRAFT ?
		_("Delete this draft from disk?") :
		_("Delete this entry from server?");

	g_assert(type != ENTRY_NEW);

	if (!lj_confirm(GTK_WINDOW(ljw),
				_("Delete"), confirm_text))
		return;

	if (type == ENTRY_DRAFT) {
		delete_draft(ljw);
	} else {
		delete_server_entry(ljw);
	}
}

static void
lj_autosave_delete(void) {
	char *path;
	path = g_build_filename(app.conf_dir, "draft", NULL);
	if (unlink(path) < 0) {
		/* FIXME handle error. */
	}
	g_free(path);
}

static void 
do_logout(LJWin *ljw) {
	string_replace(&conf.usejournal, NULL);
	usejournal_update(USEJOURNAL(ljw->usejournal));
	cfmgr_set_state(ljw->cfmgr, CF_DISABLED);

	conf.loginok = FALSE;
	menu_update(ljw, NULL);
	usejournal_update(USEJOURNAL(ljw->usejournal));
}

void 
lj_do_loginout(LJWin *ljw) {
	if (conf.loginok)
		do_logout(ljw);
	if (app.autolog) {
		if (app.skiplogin) {
			conf.loginok = TRUE;
			menu_update(ljw, NULL);
		} else {
			login_run(GTK_WINDOW(ljw), conf_cur_user(), ljw);
		}
		app.skiplogin = FALSE;
		app.autolog = FALSE;
	} else {
		login_dlg_run(ljw);
	}

	usejournal_update(USEJOURNAL(ljw->usejournal));
	if (conf.loginok) {
		gtk_widget_show_all(cfindicator_box_get(ljw->cfi_main));

		cfmgr_mask_set(app.cfmgr, conf_cur_user()->cfmask);
		if (conf.options.cfautostart)
			cfmgr_set_state(ljw->cfmgr, CF_ON);
	} else {
		gtk_widget_hide_all(cfindicator_box_get(ljw->cfi_main));
	}
	logjam_doc_user_changed(ljw->doc);
	metamgr_login_changed(METAMGR(ljw->metamgr));
	gtk_widget_grab_focus(ljw->eentry);
}

static gboolean
lj_remote_change_user_cb(LJWin *ljw,
		char *username, gboolean skiplogin, GError **err) {
	if (username) {
		User *u;
		u = server_get_user_by_username(conf_cur_server(), username);
		if (!u)
			return FALSE;
		conf_cur_user() = u;

		if (skiplogin) {
			app.autolog = TRUE;
			app.skiplogin = TRUE;
		}
		lj_do_loginout(ljw);
	}
	gtk_window_present(GTK_WINDOW(ljw));
	return TRUE;
}

void
lj_save_autosave(LJWin *ljw) {
	Entry *entry;
	char *path;

	path = g_build_filename(app.conf_dir, "draft", NULL);
	entry = logjam_doc_get_entry(ljw->doc);
	if (!entry_to_xml_file(entry, path, NULL)) {
		/* FIXME handle error. */
	}
	entry_free(entry);
	g_free(path);
}
	
gboolean
lj_save_autosave_cb(LJWin *ljw) {
	if (!logjam_doc_get_dirty(ljw->doc))
		return TRUE;

	lj_save_autosave(ljw);

	return TRUE; /* perpetuate timeout */
}

void 
lj_open_entry(LJWin *ljw) {
	GtkWidget *filesel;
	GtkWidget *hbox, *filetype;
	GtkWidget *menu, *item;
	static struct {
		char *label;
		EntryFileType type;
	} filetypes[] = {
		{ N_("Autodetect"),   ENTRY_FILE_AUTODETECT },
		{ N_("XML"),          ENTRY_FILE_XML        },
		{ N_("RFC822-Style"), ENTRY_FILE_RFC822     },
		{ N_("Plain Text"),   ENTRY_FILE_PLAIN      },
		{ NULL                                  },
	}, *ft;

	if (!lj_confirm_lose_entry(ljw)) return;
	
	filesel = lj_filesel_new(_("Open Entry"), GTK_WINDOW(ljw));

	filetype = gtk_option_menu_new();
	menu = gtk_menu_new();
	for (ft = filetypes; ft->label; ft++) {
		item = gtk_menu_item_new_with_label(_(ft->label));
		gtk_widget_show(item);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
	}
	gtk_option_menu_set_menu(GTK_OPTION_MENU(filetype), menu);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new(_("File Type:")),
			FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), filetype, TRUE, TRUE, 0);
	gtk_widget_show_all(hbox);

	gtk_box_pack_start(GTK_BOX(GTK_FILE_SELECTION(filesel)->main_vbox),
			hbox, FALSE, FALSE, 0);
	
	while (gtk_dialog_run(GTK_DIALOG(filesel)) == GTK_RESPONSE_OK) {
		const gchar *filename;
		GError *err = NULL;
		int op = gtk_option_menu_get_history(GTK_OPTION_MENU(filetype));
		EntryFileType type = filetypes[op].type;

		filename = gtk_file_selection_get_filename(
				GTK_FILE_SELECTION(filesel));

		if (!logjam_doc_load_file(ljw->doc, filename, type, &err)) {
			lj_warning(GTK_WINDOW(filesel), err->message);
			g_error_free(err);
		} else {
			break;
		}
	}
	gtk_widget_destroy(filesel);
}

void
lj_open_draft(LJWin *ljw) {
	DraftStore *ds;
	Entry *e;

	if (!lj_confirm_lose_entry(ljw)) return;

	ds = draft_store_new_from_current();

	e = draft_store_ui_select(ds, GTK_WINDOW(ljw));
	if (e) {
		logjam_doc_load_draft(ljw->doc, e);
		entry_free(e);
	}

	draft_store_free(ds);
}

#ifdef USE_STRUCTUREDTEXT
char* structuredtext_to_html(const char *src, GError **err);
char* html_to_structuredtext(const char *src, GError **err);
typedef char* (*TextStyleFunc)(const char *src, GError **err);
static TextStyleFunc to_html[] = {
	NULL,
	structuredtext_to_html
};
static TextStyleFunc from_html[] = {
	NULL,
	html_to_structuredtext
};

static gboolean
lj_change_textstyle(LJWin *ljw, TextStyle newstyle, GError **err) {
	GtkTextBuffer *buffer;
	GtkTextIter start, end;
	char *src;
	char *html, *result;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(ljw->eentry));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
	src = gtk_text_buffer_get_text(buffer, &start, &end, TRUE);

	if (to_html[ljw->textstyle])
		html = to_html[ljw->textstyle](src, err);
	else
		html = src;

	if (html == NULL) {
		g_free(src);
		return FALSE;
	}

	if (from_html[newstyle])
		result = from_html[newstyle](src, err);
	else
		result = html;

	if (result == NULL) {
		g_free(src);
		if (html != src)
			g_free(html);
		return FALSE;
	}
	
	ljw->textstyle = newstyle;
	gtk_text_buffer_set_text(buffer, result, -1);

	g_free(src);
	if (src != html)
		g_free(html);
	if (html != result)
		g_free(result);
	return TRUE;
}

static void
textstyle_changed_cb(GtkOptionMenu *om, LJWin *ljw) {
	TextStyle newstyle = gtk_option_menu_get_history(om);
	GError *err = NULL;

	if (newstyle == ljw->textstyle)
		return;

	if (!lj_change_textstyle(ljw, newstyle, &err)) {
		lj_message(GTK_WIDGET(ljw), LJ_MSG_ERROR, FALSE,
				_("Error Changing Text Format"), "%s", err->message);
		g_error_free(err);
		gtk_option_menu_set_history(om, ljw->textstyle);
	}
}

static void
make_textmods(LJWin *ljw) {
	GtkWidget *hbox;
	GtkWidget *omenu, *menu, *item;
	GtkRequisition req;

	omenu = gtk_option_menu_new();
	menu = gtk_menu_new();
	item = gtk_menu_item_new_with_label(_("Normal"));
	gtk_widget_show(item);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);

	item = gtk_menu_item_new_with_label(_("Structured"));
	gtk_widget_show(item);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);

	/*item = gtk_menu_item_new_with_label("Formatted");
	gtk_widget_set_sensitive(item, FALSE);
	gtk_widget_show(item);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);*/

	gtk_option_menu_set_menu(GTK_OPTION_MENU(omenu), menu);

	g_signal_connect(G_OBJECT(omenu), "changed",
			G_CALLBACK(textstyle_changed_cb), ljw);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new(_("Entry Format:")),
			FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), omenu, FALSE, FALSE, 0);

	gtk_widget_size_request(omenu, &req);

	gtk_text_view_set_border_window_size(GTK_TEXT_VIEW(ljw->eentry),
			GTK_TEXT_WINDOW_TOP, req.height+10);
	gtk_text_view_add_child_in_window(GTK_TEXT_VIEW(ljw->eentry),
			hbox, GTK_TEXT_WINDOW_TOP, 5, 5);
}
#endif /* USE_STRUCTUREDTEXT */

static GtkWidget*
make_main_entry(LJWin *ljw) {
	ljw->eentry = gtk_text_view_new_with_buffer(
			logjam_doc_get_text_buffer(ljw->doc));
	gtk_text_view_set_editable(GTK_TEXT_VIEW(ljw->eentry), TRUE);
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(ljw->eentry), GTK_WRAP_WORD);

	undo_manager_attach(ljw->um, ljw->eentry);

	g_signal_connect(G_OBJECT(ljw->eentry), "populate-popup",
		G_CALLBACK(populate_entry_popup), ljw);

	g_signal_connect(G_OBJECT(ljw->um), "can-undo",
		G_CALLBACK(can_undo_cb), ljw);
	g_signal_connect(G_OBJECT(ljw->um), "can-redo",
		G_CALLBACK(can_redo_cb), ljw);

#ifdef USE_STRUCTUREDTEXT
	make_textmods(ljw);
#endif /* USE_STRUCTUREDTEXT */
#ifdef HAVE_GTKSPELL
	if (conf.options.usespellcheck) {
		GError *err = NULL;
		if (gtkspell_new_attach(GTK_TEXT_VIEW(ljw->eentry), conf.spell_language, &err) == NULL) {
			lj_warning(GTK_WINDOW(ljw), _("GtkSpell error: %s"), err->message);
			g_error_free(err);
		}
	}
#endif /* HAVE_GTKSPELL */

	/* fixme gtk2; this would submit the entry when you hit ctl-enter
	g_signal_connect(G_OBJECT(ljw->eentry), "activate",
		G_CALLBACK(submit_entry), ljw);*/
	if (conf.uifont)
		lj_widget_set_font(ljw->eentry, conf.uifont);

	return scroll_wrap(ljw->eentry);
}

static GtkWidget*
make_action_bar(LJWin *ljw) {
	ljw->actionbox = gtk_hbox_new(FALSE, 5); 

	ljw->cfi_main = cfindicator_new(ljw->cfmgr, CF_MAIN, GTK_WINDOW(ljw));
	cfmgr_set_state(ljw->cfmgr, CF_DISABLED);
	gtk_box_pack_start(GTK_BOX(ljw->actionbox),
			cfindicator_box_get(ljw->cfi_main),
			FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(ljw->actionbox),
			gtk_label_new(NULL), TRUE, TRUE, 0);

	ljw->usejournal = usejournal_new();
	gtk_box_pack_start(GTK_BOX(ljw->actionbox),
			ljw->usejournal, FALSE, FALSE, 0);

	ljw->bdelete = gtk_button_new_from_stock(GTK_STOCK_DELETE);
	g_signal_connect(G_OBJECT(ljw->bdelete), "clicked",
			G_CALLBACK(delete_cb), ljw);
	gtk_box_pack_start(GTK_BOX(ljw->actionbox), ljw->bdelete, FALSE, FALSE, 0);

	ljw->baction = gtk_button_new();
	ljw->laction = gtk_label_new(NULL);
	g_signal_connect(G_OBJECT(ljw->baction), "clicked",
			G_CALLBACK(action_cb), ljw);
	gtk_container_add(GTK_CONTAINER(ljw->baction), ljw->laction);
	gtk_box_pack_start(GTK_BOX(ljw->actionbox), ljw->baction, FALSE, FALSE, 0);

	return ljw->actionbox;
}

static GtkWidget*
make_app_contents(LJWin *ljw) {
	GtkWidget *vbox;

	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);

	/* the cfmgr will very soon keep track of the
	 * default cf indicator in the submit bar. */
	ljw->cfmgr = cfmgr_new();
	cfmgr_parent_set(ljw->cfmgr, GTK_WINDOW(ljw));
	/* (see comment in checkfriends.c about this circular reference.) */

	ljw->metamgr = metamgr_new();
	gtk_box_pack_start(GTK_BOX(vbox), ljw->metamgr, FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(vbox), 
			make_main_entry(ljw), TRUE, TRUE, 0);
	gtk_box_pack_end(GTK_BOX(vbox), 
			make_action_bar(ljw), FALSE, FALSE, 0);
	
	return vbox;
}

void
lj_autosave_init(LJWin* ljw) {
	if (app.autosave || !conf.options.autosave || !app.autosave_ok)
		return;

	app.autosave = g_timeout_add(LJ_AUTOSAVE_INTERVAL,
			(GSourceFunc)lj_save_autosave_cb, ljw);
}

void
lj_autosave_stop(LJWin* ljw) {
	if (!app.autosave)
		return;
	lj_autosave_delete();
	g_source_remove(app.autosave);
	app.autosave = 0;
}

/* load a saved autosave if there is one and it's appropriate to do so */
static Entry*
lj_load_autosave(LJWin* ljw) {
	if (conf.options.autosave && app.autoloaddraft_ok) {
		GError *err = NULL;
		char *path;
		Entry *entry;

		path = g_build_filename(app.conf_dir, "draft", NULL);

		if (!g_file_test(path, G_FILE_TEST_EXISTS)) {
			g_free(path);
			return FALSE;
		}

		entry = entry_new_from_filename_xml(path, &err);
		if (entry == NULL) {
			lj_warning(GTK_WINDOW(ljw), _("Error loading draft: %s."),
					err->message);
			g_error_free(err);
		}
		g_free(path);
		return entry;
	}
	return NULL; /* didn't read draft */
}

void
lj_quit(LJWin *ljw) {
	if (lj_confirm_lose_entry(ljw))
		gtk_main_quit();
}

static gboolean
delete_event_cb(LJWin *ljw) {
	lj_quit(ljw);
	return TRUE; /* don't ever let this delete the window; quit will do it. */
}

/* gtk stuff */
static GType
ljwin_get_type(void) {
	static GType ljw_type = 0;
	if (!ljw_type) {
		const GTypeInfo ljw_info = {
			sizeof (GtkWindowClass),
			NULL,
			NULL,
			NULL,
			NULL,
			NULL,
			sizeof (LJWin),
			0,
			//(GInstanceInitFunc) ljwin_init, /* no init func needed since */
			NULL,     /* GTK_WINDOW_TOPLEVEL is the default GtkWindow:type */
		};
		ljw_type = g_type_register_static(GTK_TYPE_WINDOW,
				"LJWin", &ljw_info, 0);
	}
	return ljw_type;
}

#define LJWIN(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj), ljwin_get_type(), LJWin))

static GtkWidget*
ljwin_new(void) {
	LJWin *ljw = LJWIN(g_object_new(ljwin_get_type(), NULL));
	ljw->um = undo_manager_new();
	/*FIXME: add the rest of the changable entry widgets */
	return GTK_WIDGET(ljw);
}

void 
lj_run(LogJamDoc *doc) {
	GtkWidget *vbox;
	Entry *draftentry;

	LJWin *ljw = LJWIN(ljwin_new());
	gtk_window_set_default_size(GTK_WINDOW(ljw), 400, 300);
	geometry_tie(GTK_WIDGET(ljw), GEOM_MAIN);

	lj_new_doc(ljw, doc);

	app.tooltips = gtk_tooltips_new();
	gtk_tooltips_enable(app.tooltips);

	g_signal_connect(G_OBJECT(ljw), "delete-event",
			G_CALLBACK(delete_event_cb), NULL);
	g_signal_connect_swapped(G_OBJECT(app.remote), "present",
			G_CALLBACK(gtk_window_present), ljw);
	g_signal_connect_swapped(G_OBJECT(app.remote), "change_user",
			G_CALLBACK(lj_remote_change_user_cb), ljw);

	vbox = gtk_vbox_new(FALSE, 0); 

	gtk_box_pack_start(GTK_BOX(vbox), 
			menu_make_bar(ljw), FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(vbox), make_app_contents(ljw), 
			TRUE, TRUE, 0);

	gtk_container_add(GTK_CONTAINER(ljw), vbox);

	lj_autosave_init(ljw);

	gtk_widget_show_all(vbox);
	lj_update_actions(ljw);
	gtk_widget_hide_all(cfindicator_box_get(ljw->cfi_main));

	menu_update(ljw, NULL);

	gtk_widget_show(GTK_WIDGET(ljw));

	/* suck a bunch of events in. */
	while (gtk_events_pending())
		gtk_main_iteration();

	if (!conf.options.allowmultipleinstances) {
		GError *err = NULL;
		if (!logjam_remote_listen(app.remote, &err)) {
			if (err) {
				lj_warning(GTK_WINDOW(ljw),
						_("Error initializing remote command socket: %s."),
						err->message);
				g_error_free(err);
			}
		}
	}

	draftentry = lj_load_autosave(ljw);
	if (draftentry) {
		if (lj_confirm(GTK_WINDOW(ljw), _("Autosaved Draft Found"),
					_("An autosaved draft was found, possibly from a previous run of LogJam that crashed.  "
					  "Would you like to recover it?"))) {
			logjam_doc_load_entry(ljw->doc, draftentry);
		}
		entry_free(draftentry);
	}

#ifndef G_OS_WIN32
	cf_update_dock(ljw->cfmgr, GTK_WINDOW(ljw));
#endif

	/* to prevent flicker, make this GtkWindow immediately after
	 * pending events have been handled, and immediately handle its
	 * own event afterwards.
	 *
	 * XXX: Should make_cf_float() have its own event-sucking loop? */
	cf_update_float(ljw->cfmgr, GTK_WINDOW(ljw));
	while (gtk_events_pending())
		gtk_main_iteration();

	lj_do_loginout(ljw);
	gtk_main();

	lj_autosave_delete();
}
