/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include "config.h"

#include <glib.h>
#include "util.h"
#include "draftstore.h"
#include "logjam-doc.h"

#include "smartquotes.h"

enum {
	TITLE_CHANGED,
	ENTRY_CHANGED,
	UPDATE_ENTRY,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_DIRTY
};

struct _LogJamDoc {
	GObject parent;

	/* for drafts, the current entry is tracked through entry->itemid.
	 * for files, we need to keep the filename around. */
	char *filename;
	/* we only save as xml, so we need to remember whether they opened
	 * a non-xml file. */
	gboolean filename_not_xml;

	/* the convention is:
	 *   if there's a filename, we're editing a file.
	 *   if itemid is less than zero, we're editing a draft.
	 *   otherwise, it's an untitled/unattached doc.
	 */

	gboolean dirty;

	Entry *entry;
	GtkTextBuffer *buffer;
	guint buffer_signal;
};

struct _LogJamDocClass {
	GObjectClass parent_class;
	void (*title_changed)(LogJamDoc *doc);
	void (*entry_changed)(LogJamDoc *doc, Entry *entry);
	void (*update_entry)(LogJamDoc *doc, Entry *entry);
};

static void logjam_doc_update_entry(LogJamDoc *doc);
static void logjam_doc_set_property(GObject *object, guint prop_id,
                        const GValue *value, GParamSpec *pspec);
static void logjam_doc_get_property(GObject *object, guint prop_id,
                        GValue *value, GParamSpec *pspec);

static guint signals[LAST_SIGNAL] = { 0 };

static void
logjam_doc_class_init(gpointer klass, gpointer class_data) {
	GObjectClass *gclass = G_OBJECT_CLASS(klass);
	//LogJamDocClass *dclass = LOGJAM_DOC_CLASS(klass);

	gclass->set_property = logjam_doc_set_property;
	gclass->get_property = logjam_doc_get_property;

	signals[TITLE_CHANGED] = g_signal_new("title_changed",
			G_OBJECT_CLASS_TYPE(gclass),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET(LogJamDocClass, title_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);

	signals[ENTRY_CHANGED] = g_signal_new("entry_changed",
			LOGJAM_TYPE_DOC,
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET(LogJamDocClass, entry_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__POINTER,
			G_TYPE_NONE, 1, G_TYPE_POINTER);

	signals[UPDATE_ENTRY] = g_signal_new("update_entry",
			LOGJAM_TYPE_DOC,
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET(LogJamDocClass, update_entry),
			NULL, NULL,
			g_cclosure_marshal_VOID__POINTER,
			G_TYPE_NONE, 1, G_TYPE_POINTER);

	g_object_class_install_property(gclass, PROP_DIRTY,
			g_param_spec_boolean("dirty",
				_("Is dirty"),
				_("If TRUE, the document has changed since it was last saved."),
				FALSE, G_PARAM_READWRITE));
}

static void
buffer_changed_cb(GtkTextBuffer *buf, LogJamDoc *doc) {
	if (gtk_text_buffer_get_modified(buf))
		logjam_doc_set_dirty(doc, TRUE);
}

static void
logjam_doc_init(GTypeInstance *inst, gpointer g_class) {
	LogJamDoc *doc = LOGJAM_DOC(inst);
	doc->buffer = gtk_text_buffer_new(NULL);
	doc->buffer_signal = g_signal_connect(G_OBJECT(doc->buffer),
			"modified-changed",
			G_CALLBACK(buffer_changed_cb), doc);
	doc->entry = entry_new();
	doc->dirty = FALSE;
	gtk_text_buffer_set_modified(doc->buffer, FALSE);
	if (conf.options.smartquotes)
		smartquotes_attach(doc->buffer);
}

GType
logjam_doc_get_type() {
	static GType doc_type = 0;
	if (!doc_type) {
		static const GTypeInfo doc_info = {
			sizeof(LogJamDocClass),
			NULL,
			NULL,
			logjam_doc_class_init,
			NULL,
			NULL,
			sizeof(LogJamDoc),
			0,
			logjam_doc_init
		};
		doc_type = g_type_register_static(G_TYPE_OBJECT, "LogJamDoc",
				&doc_info, 0);
	}
	return doc_type;
}

static void
logjam_doc_set_property(GObject *object, guint prop_id,
                        const GValue *value, GParamSpec *pspec)
{
	/* XXX */
}
static void
logjam_doc_get_property(GObject *object, guint prop_id,
                        GValue *value, GParamSpec *pspec)
{
	/* XXX */
}

LogJamDoc*
logjam_doc_new() {
	return g_object_new(logjam_doc_get_type(), NULL);
}

GtkTextBuffer*
logjam_doc_get_text_buffer(LogJamDoc *doc) {
	return doc->buffer;
}

void
logjam_doc_set_dirty(LogJamDoc *doc, gboolean dirty) {
	if (doc->dirty == dirty)
		return;
	doc->dirty = dirty;
	if (doc->dirty == FALSE)
		gtk_text_buffer_set_modified(doc->buffer, FALSE);
	g_object_notify(G_OBJECT(doc), "dirty");
}

gboolean
logjam_doc_get_dirty(LogJamDoc *doc) {
	return doc->dirty;
}

Security
logjam_doc_get_security(LogJamDoc *doc) {
	return doc->entry->security;
}
void
logjam_doc_set_security(LogJamDoc *doc, Security *sec) {
	doc->entry->security = *sec;
}
void
logjam_doc_set_comments(LogJamDoc *doc, CommentsType type) {
	doc->entry->comments = type;
}

void
logjam_doc_set_backdated(LogJamDoc *doc, gboolean backdated) {
	doc->entry->backdated = backdated;
}

void
logjam_doc_set_preformatted(LogJamDoc *doc, gboolean preformatted) {
	doc->entry->preformatted = preformatted;
}

void
logjam_doc_set_pickeyword(LogJamDoc *doc, const char *keyword) {
	string_replace(&doc->entry->pickeyword, keyword ? g_strdup(keyword) : NULL);
}

static void
set_entry(LogJamDoc *doc, Entry *entry) {
	GtkTextIter start, end;

	entry_free(doc->entry);
	if (entry)
		doc->entry = entry;
	else
		entry = doc->entry = entry_new();

	/* block the buffer signal so we don't rapidly flip to dirty and back. */
	g_signal_handler_block(doc->buffer, doc->buffer_signal);
	gtk_text_buffer_get_bounds(doc->buffer, &start, &end);
	gtk_text_buffer_delete(doc->buffer, &start, &end);
	if (entry->event)
		gtk_text_buffer_insert(doc->buffer, &start, entry->event, -1);
	g_signal_handler_unblock(doc->buffer, doc->buffer_signal);

	logjam_doc_set_dirty(doc, FALSE);
	/* since the buffer signal was blocked, we need to do this manually. */
	gtk_text_buffer_set_modified(doc->buffer, FALSE);
}

gboolean
logjam_doc_load_file(LogJamDoc *doc, const char *filename,
                     EntryFileType type, GError **err) {
	Entry *entry;
	entry = entry_new_from_filename(filename, &type, err);
	if (!entry)
		return FALSE;
	entry->itemid = 0;
	set_entry(doc, entry);
	string_replace(&doc->filename, g_strdup(filename));
	doc->filename_not_xml = (type != ENTRY_FILE_XML);
	g_signal_emit_by_name(doc, "title_changed");
	g_signal_emit_by_name(doc, "entry_changed", entry);
	return TRUE;
}

void
logjam_doc_load_draft(LogJamDoc *doc, Entry *entry) {
	set_entry(doc, entry_copy(entry));
	g_signal_emit_by_name(doc, "title_changed");
	g_signal_emit_by_name(doc, "entry_changed", entry);
}

void
logjam_doc_load_entry(LogJamDoc *doc, Entry *entry) {
	set_entry(doc, entry_copy(entry));
	g_signal_emit_by_name(doc, "title_changed");
	g_signal_emit_by_name(doc, "entry_changed", entry);
}

static gboolean
save_draft(Entry *entry, GError **err) {
	DraftStore *ds;

	ds = draft_store_new_from_current();

	if (entry->itemid == 0)
		entry->itemid = draft_store_find_itemid(ds);

	if (!draft_store_put_entry(ds, entry, err)) {
		draft_store_free(ds);
		return FALSE;
	}

	draft_store_free(ds);

	return TRUE;
}

gboolean
logjam_doc_would_save_over_nonxml(LogJamDoc *doc) {
	return (doc->filename && doc->filename_not_xml);
}
gboolean
logjam_doc_has_save_target(LogJamDoc *doc) {
	/* filename or a draft. */
	return (doc->filename || doc->entry->itemid < 0);
}

gboolean
logjam_doc_save(LogJamDoc *doc, GError **err) {
	logjam_doc_update_entry(doc);

	if (doc->filename) {
		if (!entry_to_xml_file(doc->entry, doc->filename, err))
			return FALSE;
	} else if (doc->entry->itemid < 0) {
		if (!save_draft(doc->entry, err))
			return FALSE;
	} else {
		/* XXX we should prompt for a save as here. */
		g_set_error(err, 0, 0, "Not yet implemented.");
		return FALSE;
	}

	g_signal_emit_by_name(doc, "title_changed");
	logjam_doc_set_dirty(doc, FALSE);

	return TRUE;
}

gboolean
logjam_doc_save_as_file(LogJamDoc *doc, const char *filename, GError **err) {
	Entry *entry;

	entry = logjam_doc_get_entry(doc);

	/* always save to files without an itemid. */
	entry->itemid = 0;
	if (!entry_to_xml_file(entry, filename, err)) {
		entry_free(entry);
		return FALSE;
	}
	entry_free(entry);

	doc->entry->itemid = 0;
	string_replace(&doc->filename, g_strdup(filename));

	g_signal_emit_by_name(doc, "title_changed");
	logjam_doc_set_dirty(doc, FALSE);

	return TRUE;
}

gboolean
logjam_doc_save_as_draft(LogJamDoc *doc, const char *title, GError **err) {
	Entry *entry;
	
	entry = logjam_doc_get_entry(doc);
	string_replace(&entry->subject, g_strdup(title));
	entry->itemid = 0;
	if (!save_draft(entry, err)) {
		entry_free(entry);
		return FALSE;
	}
	/* retrieve the new draft id. */
	doc->entry->itemid = entry->itemid;
	entry_free(entry);

	string_replace(&doc->filename, NULL);

	g_signal_emit_by_name(doc, "title_changed");
	logjam_doc_set_dirty(doc, FALSE);

	return TRUE;
}

char*
logjam_doc_get_title(LogJamDoc *doc) {
	if (doc->filename) {
		const gchar *homedir = g_get_home_dir();
		int homelen = strlen(homedir);
		if (strncmp(homedir, doc->filename, homelen) == 0)
			return g_build_filename("~", doc->filename+homelen, NULL);
		else
			return g_strdup(doc->filename);
	} else if (doc->entry->itemid < 0) {
		return g_strdup_printf(_("Draft %d"), -doc->entry->itemid);
	} else if (doc->entry->itemid > 0) {
		return g_strdup_printf(_("Entry %d"), doc->entry->itemid);
	} else {
		return g_strdup(_("New Entry"));
	}
}

char*
logjam_doc_get_draftname(LogJamDoc *doc) {
	logjam_doc_update_entry(doc);
	if (doc->entry->subject)
		return g_strdup(doc->entry->subject);
	return NULL;
}

static void
logjam_doc_update_entry(LogJamDoc *doc) {
	/* XXX this is sorta weird; we emit a signal to get our internal version
	 * of the entry in sync with the state of the widgets.  this is important
	 * in decoupling the widgets from the document, but it feels a little
	 * backwards. */
	g_signal_emit_by_name(doc, "update_entry", doc->entry);
}

Entry*
logjam_doc_get_entry(LogJamDoc *doc) {
	logjam_doc_update_entry(doc);
	return entry_copy(doc->entry);
}

gboolean
logjam_doc_insert_file(LogJamDoc *doc, const char *filename, const char *encoding, GError **err) {
	char *text;
	gsize len;

	if (!g_file_get_contents(filename, &text, &len, err))
		return FALSE;

	if (strcmp(encoding, "UTF-8") == 0) {
		const gchar *end;
		if (!g_utf8_validate(text, len, &end)) {
			g_set_error(err, 0, 0,
					_("Invalid UTF-8 starting at byte %d."), end-text);
			g_free(text);
			return FALSE;
		}
		gtk_text_buffer_insert_at_cursor(doc->buffer, text, -1);
	} else {
		char *newtext;
		newtext = g_convert(text, -1, "UTF-8", encoding, NULL, NULL, err);
		if (!newtext) {
			g_free(text);
			return FALSE;
		}
		gtk_text_buffer_insert_at_cursor(doc->buffer, newtext, -1);
		g_free(newtext);
	}
	g_free(text);
	return TRUE;
}

int
logjam_doc_get_flags(LogJamDoc *doc) {
	EntryType type = logjam_doc_get_entry_type(doc);
	int flags = 0;

	switch (type) {
		case ENTRY_DRAFT:
			flags = LOGJAM_DOC_CAN_DELETE | LOGJAM_DOC_CAN_SUBMIT;
			break;
		case ENTRY_NEW:
			flags = LOGJAM_DOC_CAN_SUBMIT;
			break;
		case ENTRY_SERVER:
			flags = LOGJAM_DOC_CAN_DELETE | LOGJAM_DOC_CAN_SAVE;
			break;
	}
	return flags;
}

EntryType
logjam_doc_get_entry_type(LogJamDoc *doc) {
	gint itemid = logjam_doc_get_entry_itemid(doc);
	return itemid < 0 ? ENTRY_DRAFT :
		itemid == 0 ? ENTRY_NEW :
		ENTRY_SERVER;
}

gint
logjam_doc_get_entry_itemid(LogJamDoc *doc) {
	return doc->entry->itemid;
}

void
logjam_doc_user_changed(LogJamDoc *doc) {
	/* if they were editing an old entry or a draft,
	 * that itemid is no longer relevant for the current user. */
	logjam_doc_update_entry(doc);
	doc->entry->itemid = 0;
	g_signal_emit_by_name(doc, "title_changed");
	g_signal_emit_by_name(doc, "entry_changed", doc->entry);
}

