/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include "config.h"
#include <gtk/gtk.h>
#include <glib.h>

#include <unistd.h>
#include <errno.h>

#include <curl/curl.h>

#include <sys/types.h>
#include <sys/wait.h>  /* waitpid */
#include <signal.h>    /* kill */

#include "protocol.h"
#include "util.h"
#include "conf.h"
#include "network.h"
#include "network-internal.h"

typedef struct {
	CURL *curl;

	GString *response;

	NetStatusCallback user_cb;
	gpointer user_data;
} RequestData;

typedef struct {
	GString *response;

	GError **err;

	guint pipe_tag;
	int pipefds[2];

	pid_t pid;

	NetStatusCallback user_cb;
	gpointer user_data;
} ForkData;

static int
pipe_write(int pipe, NetStatusType type, int len, void *data) {
	char t = type;
	if (write(pipe, &t, 1) < 1)
		return -1;
	if (write(pipe, &len, sizeof(int)) < sizeof(int))
		return -1;
	if (data) {
		if (write(pipe, data, len) < len)
			return -1;
	}
	return 0;
}

static size_t 
curlwrite_cb(void *ptr, size_t size, size_t nmemb, void *data) {
	RequestData *requestdata = data;
	double contentlength;
	
	g_string_append_len(requestdata->response, ptr, size*nmemb);

	curl_easy_getinfo(requestdata->curl, 
			CURLINFO_CONTENT_LENGTH_DOWNLOAD, &contentlength);

	if (requestdata->user_cb) {
		NetStatusProgress progress = {0};
		progress.current = requestdata->response->len;
		progress.total = (int)contentlength;

		requestdata->user_cb(NET_STATUS_PROGRESS, &progress, requestdata->user_data);
	}

	return size*nmemb;
}

/*
 * see comment in run_curl_request().
static int
curl_progress_cb(void *cp, size_t dlt, size_t dln, size_t ult, size_t uln) {
	g_print("progress: %d %d %d %d\n", dlt, dln, ult, uln);
	return 0;
}*/

GString*
net_post_blocking(GString *post,
                  NetStatusCallback cb, gpointer data,
                  GError **err) {
	CURL *curl;
	CURLcode curlres;
	char urlreq[1024], proxyuserpass[1024];
	char errorbuf[CURL_ERROR_SIZE];
	RequestData requestdata = {0};

	curl = curl_easy_init();
	if (curl == NULL) {
		g_set_error(err, 0, 0, "Unable to intialize CURL");
		return NULL;
	}

	curl_easy_setopt(curl, CURLOPT_VERBOSE, conf.options.netdump != 0);
	curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);
	curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, errorbuf);

	/*
	 * curl's progress function is mostly useless; we instead track writes.
	 *
	 * curl_easy_setopt(curl, CURLOPT_PROGRESSFUNCTION, curl_progress_cb);
	 * curl_easy_setopt(curl, CURLOPT_PROGRESSDATA, nr);
	 */
	
	g_snprintf(urlreq, sizeof(urlreq), 
			"%s/interface/flat", conf_cur_server()->url);
	curl_easy_setopt(curl, CURLOPT_URL, urlreq);

	if (conf.options.useproxy) {
		curl_easy_setopt(curl, CURLOPT_PROXY, conf.proxy);
		if (conf.options.useproxyauth) {
			g_snprintf(proxyuserpass, sizeof(proxyuserpass), "%s:%s", 
					conf.proxyuser, conf.proxypass);
			curl_easy_setopt(curl, CURLOPT_PROXYUSERPWD, proxyuserpass);
		}
	}

	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post->str);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, post->len);
	curl_easy_setopt(curl, CURLOPT_POST, 1);

	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curlwrite_cb);
	requestdata.curl = curl;
	requestdata.response = g_string_sized_new(READ_BLOCK_SIZE);
	requestdata.user_cb = cb;
	requestdata.user_data = data;
	curl_easy_setopt(curl, CURLOPT_FILE, &requestdata);
	
	if (conf.options.netdump) 
		fprintf(stderr, _("Request: [%s]\n"), post->str);

	curlres = curl_easy_perform(curl);
	curl_easy_cleanup(curl);

	if (curlres != CURLE_OK) {
		g_set_error(err, 0, 0, "cURL error: %s.", errorbuf);
		g_string_free(requestdata.response, TRUE);
		return NULL;
	}

	if (conf.options.netdump) 
		fprintf(stderr, _("Response: [%s]\n"), requestdata.response->str);

	return requestdata.response;
}

static void
fork_cb(NetStatusType status, gpointer statusdata, gpointer data) {
	ForkData *forkdata = data;
	if (status == NET_STATUS_PROGRESS)
		pipe_write(forkdata->pipefds[1], NET_STATUS_PROGRESS,
				sizeof(NetStatusProgress), statusdata);
}

static void
pipe_cb(ForkData *forkdata, gint pipe, GdkInputCondition cond) {
	char t;
	int len;
	int rlen = 0;
	char *buf;
	NetStatusProgress progress;

	len = read(pipe, &t, 1);
	if (len == 0)
		g_print("Pipe unexpectedly closed.\n");
	else if (len < 0)
		perror("read");
	if (read(pipe, &len, sizeof(int)) < sizeof(int))
		return;

	switch (t) {
		case NET_STATUS_SUCCESS:
			buf = g_new0(char, len+1);
			do {
				rlen += read(pipe, buf+rlen, len-rlen);
			} while (rlen < len);
			buf[len] = 0;
			waitpid(forkdata->pid, NULL, 0);
			forkdata->pid = 0;
			close(pipe);
			forkdata->response = g_string_new_len(buf, len);
			g_free(buf);
			gtk_main_quit();
			break;
		case NET_STATUS_ERROR:
			buf = g_new0(char, len+1);
			do {
				rlen += read(pipe, buf+rlen, len-rlen);
			} while (rlen < len);
			buf[len] = 0;
			waitpid(forkdata->pid, NULL, 0);
			forkdata->pid = 0;
			close(pipe);
			g_set_error(forkdata->err, 0, 0, buf);
			g_free(buf);
			gtk_main_quit();
			break;
		case NET_STATUS_PROGRESS:
			read(pipe, &progress, sizeof(NetStatusProgress));
			if (forkdata->user_cb)
				forkdata->user_cb(NET_STATUS_PROGRESS, &progress, forkdata->user_data);
			break;
		default:
			g_warning("pipe_cb: unhandled status %d.\n", t);
	}
}

void
net_mainloop_cancel(NetMainloopHandle handle) {
	ForkData *forkdata = (ForkData*)handle;

	if (forkdata->pid > 0) {
		kill(forkdata->pid, SIGKILL);
		gtk_input_remove(forkdata->pipe_tag);
		forkdata->pipe_tag = 0;
		waitpid(forkdata->pid, NULL, 0);
		forkdata->pid = 0;

		close(forkdata->pipefds[0]);
		close(forkdata->pipefds[1]);
		g_set_error(forkdata->err, 0, NET_ERROR_CANCELLED, _("Cancelled."));
	}
	gtk_main_quit();
}

GString*
net_post_mainloop(GString *post,
                  NetStatusCallback cb, gpointer data,
                  GError **err) {
	ForkData forkdata = {0};

	forkdata.err = err;
	forkdata.user_cb = cb;
	forkdata.user_data = data;

	/* fork, run the request, then pipe the data back out of the fork. */
	if (pipe(forkdata.pipefds) < 0) {
		g_set_error(err, 0, 0,
				"Error creating pipe (pipe(): %s).", g_strerror(errno));
		return NULL;
	}
	forkdata.pid = fork();
	if (forkdata.pid < 0) {
		g_set_error(err, 0, 0,
				"Error forking (fork(): %s).", g_strerror(errno));
		return NULL;
	} else if (forkdata.pid == 0) { /* child. */
		GString *response;
		GError *err = NULL;

		response = net_post_blocking(post, fork_cb, &forkdata, &err);
		if (response == NULL) {
			int len = strlen(err->message);
			pipe_write(forkdata.pipefds[1], NET_STATUS_ERROR, len, err->message);
			g_error_free(err);
		} else {
			pipe_write(forkdata.pipefds[1], NET_STATUS_SUCCESS, response->len, response->str);
			g_string_free(response, TRUE);
		}

		close(forkdata.pipefds[0]);
		close(forkdata.pipefds[1]);

		_exit(0);
	} 
	/* otherwise, we're the parent. */
	forkdata.pipe_tag = gtk_input_add_full(forkdata.pipefds[0], GDK_INPUT_READ, 
			(GdkInputFunction)pipe_cb, NULL, &forkdata, NULL);
	if (cb)
		cb(NET_STATUS_BEGIN, &forkdata, data);

	gtk_main(); /* wait for the response. */
	if (forkdata.pipe_tag) {
		gtk_input_remove(forkdata.pipe_tag);
		forkdata.pipe_tag = 0;
	}
	forkdata.pid = 0;

	close(forkdata.pipefds[0]);
	close(forkdata.pipefds[1]);

	return forkdata.response;
}

