/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include <glib.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#include "util.h"
#include "protocol.h"

/* in a result, only one string is allocated,
 * and the rest of the hash table is pointers within that string.
 * we need to keep a pointer to that string around when we free the result. */
#define LOGJAM_STRING_KEY "logjam-data-string"

static GHashTable*
lj_protocol_hash_new(void) {
	return g_hash_table_new(g_str_hash, g_str_equal);
}

/* retrieve a pointer to the current line, modifying src in place. */
static char *
get_line_inplace(char *src, char **dest) {
	*dest = src;
	while (*src != 0 && *src != '\r' && *src != '\n')
		src++;

	if (*src != 0) {
		*src = 0;

		/* advance to the next line. */
		src++;
		while (*src == '\r' && *src == '\n')
			src++;
	}

	return src;
}

LJResult* 
lj_protocol_parse_response(const char *res) {
	char *src;
	char *key;
	char *val;
	LJResult *hash;
	int read_keys = 0;

	/* make our own copy so we can modify it. */
	src = g_strdup(res);
	hash = lj_protocol_hash_new();

	g_hash_table_insert(hash, LOGJAM_STRING_KEY, src);

	while (*src != 0) {
		src = get_line_inplace(src, &key);
		if (*src) {
			src = get_line_inplace(src, &val);
			g_hash_table_insert(hash, key, val);
			read_keys++;
		}
	}
	if (read_keys == 0) { /* error. */
		lj_protocol_result_free(hash);
		return NULL;
	}
	return (LJResult*)hash;
}

int 
lj_protocol_result_succeeded(LJResult *result) {
	char *error;

	if (result == NULL) return FALSE;

	error = net_result_get(result, "success");
	if (error == NULL) return FALSE;

	return (g_ascii_strcasecmp(error, "OK") == 0);
}

LJRequest*
lj_protocol_request_new(const char *mode,
		const char *username, const char *password, const char *usejournal,
		int protocolversion) {
	unsigned char buf[33];

	LJRequest *request = (LJRequest*)lj_protocol_hash_new();

	net_request_copys(request, "mode", mode);
	net_request_copys(request, "user", username);
	if (usejournal)
		net_request_copys(request, "usejournal", usejournal);
	net_request_seti(request, "ver", protocolversion);

	md5_hash(password, buf);
	net_request_copys(request, "hpassword", buf);

	return request;
}

static void
hash_append_str_encoded(gpointer key, gpointer value, gpointer data) {
	GString *string = data;
	gchar *en_key, *en_value;

	if (key == NULL) return;
	en_key = urlencode(key);
	en_value = urlencode(value);
	g_string_append_printf(string, "%s=%s&", en_key, en_value);

	g_free(en_key);
	g_free(en_value);
}

GString*
lj_protocol_request_to_string(LJRequest* request) {
	GString *str = g_string_sized_new(2048);
	g_hash_table_foreach(request, hash_append_str_encoded, str);
	g_string_truncate(str, str->len - 1);  /* snip off the trailing & */
	return str;
}

static gint
free_all_hash_cb(gpointer key, gpointer val, gpointer user_data) {
	g_free(key); g_free(val);
	return TRUE;
}

static void
lj_protocol_hash_free(GHashTable *hash) {
	g_hash_table_foreach_remove(hash, free_all_hash_cb, NULL);
	g_hash_table_destroy(hash);
}

void
lj_protocol_request_free(LJRequest *request) {
	lj_protocol_hash_free(request);
}

void
lj_protocol_result_free(LJResult *result) {
	char *str;
	if (!result) return;
	str = g_hash_table_lookup(result, LOGJAM_STRING_KEY);
	g_free(str);
	g_hash_table_destroy(result);
}

