/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include "config.h"
#include <gtk/gtk.h>

#include "util.h"
#include "journalstore.h"
#include "network.h"

typedef struct {
	struct tm time;
	int id;
} SyncItem;

typedef struct {
	GtkWidget *netwindow;
	
	JournalStore *js;
	char *lastsync;
	int synced;
	int total;

	GtkWidget *label, *progress;

	gboolean sync_complete;
	GSList *syncitems;
} SyncStatus;

static char *
get_lastsync(JournalStore *js) {
	char *filename;
	char *sync = NULL;
	gsize len;

	if (journal_store_get_invalid(js)) {
		/* the journalstore was invalid; we need to resync. */
		return NULL;
	}

	filename = conf_make_curuser_path("journal/lastsync");
	if (!g_file_get_contents(filename, &sync, &len, NULL))
		sync = NULL;
	g_free(filename);
	/*if (sync)
		g_print("read lastsync off disk as %s\n", sync);*/
	return sync;
}

static gboolean
put_lastsync(char *lastsync) {
	char *filename;
	FILE *f;

	filename = conf_make_curuser_path("journal/lastsync");
	f = fopen(filename, "w");
	if (f == NULL) {
		g_warning("can't open %s\n", filename);
		// XXX
	}
	fwrite(lastsync, strlen(lastsync), 1, f);
	fclose(f);

	g_free(filename);
	return TRUE;
}

static void
entries_foreach_cb(gpointer key, gpointer val, gpointer data) {
	SyncStatus *status = data;
	GError *err = NULL;
	if (!journal_store_put(status->js, val, &err)) {
		g_print("error putting entry.\n");
		if (err)
			g_print("err %s.\n", err->message);
	}
}

static int
run_one_getevents(SyncStatus *status) {
	NetRequest *request;
	NetResult *result;
	GHashTable *entries;
	int events_count;
	Entry *lastentry;
	int lastitemid;

	request = net_request_new("getevents");
	net_request_copys(request, "selecttype", "syncitems");
	net_request_copys(request, "lineendings", "unix");
	/* XXX if (conf.usejournal)
		net_request_copys(request, "usejournal", conf.usejournal);*/
	if (status->lastsync)
		net_request_copys(request, "lastsync", status->lastsync);
	result = net_window_run_request(NET_WINDOW(status->netwindow), request);
	net_request_free(request);

	if (!net_result_succeeded(result)) {
		net_result_free(result);
		return 0;
	}

	events_count = net_result_geti(result, "events_count");

	/*g_print("lastsync %s : got %d entries\n",
			status->lastsync ? status->lastsync : "(none)", events_count);*/
	entries = entries_new_from_result(result, events_count, &lastentry);
	lastitemid = lastentry->itemid;
	g_hash_table_foreach(entries, entries_foreach_cb, status);
	net_result_free(result);
	return lastitemid;
}

static SyncItem*
read_syncitems(SyncStatus *status, NetResult *result) {
	GSList *tail;
	int count, i;
	SyncItem *syncitem, *lastsyncitem = NULL;

	count = net_result_geti(result, "sync_count");

	tail = g_slist_last(status->syncitems);
	for (i = 1; i <= count; i++) {
		char *item = net_result_getf(result, "sync_%d_item", i);
		if (item[0] == 'L') {
			int id = atoi(item+2);
			
			/* skip duplicates. */
			if (tail && ((SyncItem*)tail->data)->id == id) {
				g_print("skipping dup.\n");
				continue;
			}

			syncitem = g_new0(SyncItem, 1);
			syncitem->id = id;
			ljdate_to_tm(net_result_getf(result, "sync_%d_time", i),
					&syncitem->time);

			lastsyncitem = syncitem;

			if (status->syncitems) {
				g_slist_append(tail, syncitem);
				tail = tail->next;
			} else {
				tail = status->syncitems = g_slist_append(tail, syncitem);
			}
		}
	}
	return lastsyncitem;
}

static gboolean
run_one_syncitems(SyncStatus *status) {
	NetRequest *request;
	NetResult *result;
	SyncItem *lastsyncitem = NULL;
	int count, total;

	request = net_request_new("syncitems");
	/* XXX if (conf.usejournal)
		net_request_copys(request, "usejournal", conf.usejournal); */
	if (status->lastsync)
		net_request_copys(request, "lastsync", status->lastsync);
	result = net_window_run_request(NET_WINDOW(status->netwindow), request);
	net_request_free(request);

	if (!net_result_succeeded(result)) {
		net_result_free(result);
		return FALSE;
	}

	total = net_result_geti(result, "sync_total");
	count = net_result_geti(result, "sync_count");

	if (!status->total) {
		status->total = total;
	}
	if (status->total > 0) {
		float fraction;
		fraction = (float)(status->total-total+count) / status->total;
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(status->progress),
				fraction);
	}
	/* printf("syncitems got %d items (%d left)\n", count, total-count); */

	if (count == total)
		status->sync_complete = TRUE;

	lastsyncitem = read_syncitems(status, result);

	net_result_free(result);

	if (lastsyncitem)
		string_replace(&status->lastsync,
				tm_to_ljdate(&lastsyncitem->time));

	return TRUE;
}

static gboolean
run_syncitems(SyncStatus *status) {
	status->lastsync = get_lastsync(status->js);

	if (status->lastsync) {
		char *text;
		text = g_strdup_printf(_("Downloading changes since\n%s..."),
				status->lastsync);
		gtk_label_set_text(GTK_LABEL(status->label), text);
		g_free(text);
	} else {
		gtk_label_set_text(GTK_LABEL(status->label),
				_("Downloading journal information..."));
	}

	do {
		if (!run_one_syncitems(status))
			return FALSE;
	} while (!status->sync_complete);
	
	/*{
		GSList *l;
		for (l = status->syncitems; l; l = l->next) {
			SyncItem *i = l->data;
			g_print("%d %d\n", i->id, i->time);
		}
	}*/

	return TRUE;
}

static gboolean
run_getevents(SyncStatus *status) {
	int lastitemid;
	GSList *l;
	SyncItem *syncitem;

	if (status->lastsync) {
		char *text;
		text = g_strdup_printf(_("Downloading journal entries since\n%s..."),
				status->lastsync);
		gtk_label_set_text(GTK_LABEL(status->label), text);
		g_free(text);
	} else {
		gtk_label_set_text(GTK_LABEL(status->label),
				_("Downloading journal..."));
	}

	do {
		lastitemid = run_one_getevents(status);
		if (lastitemid <= 0)
			break;

		/* find the last item we downloaded
		 * in our list of syncitems, tallying entries
		 * we've synced as we go. */
		syncitem = NULL;
		status->synced = 0;
		for (l = status->syncitems; l; l = l->next) {
			status->synced++; /* XXX this is incorrect if the sync returns 
			                     anything other than journal entries (such as
			                     comments, etc.). */
			syncitem = l->data;
			if (syncitem->id == lastitemid)
				break;
		}
		if (!syncitem) {
			g_warning("could not find syncitem for itemid %d!?", lastitemid);
			break;
		}

		if (status->total > 0) {
			float fraction;
			fraction = (float)status->synced / status->total;
			gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(status->progress),
					fraction);
		}

		string_replace(&status->lastsync,
				tm_to_ljdate(&syncitem->time));
		put_lastsync(status->lastsync);
	} while (l->next);

	return TRUE;
}

void
sync_run(GtkWindow *parent) {
	SyncStatus st = {0}, *status = &st;
	GError *err = NULL;

	status->netwindow = net_window_new(parent, _("Synchronizing Journal..."));
	status->label = gtk_label_new(_("Synchronizing Journal..."));
	gtk_label_set_justify(GTK_LABEL(status->label), GTK_JUSTIFY_CENTER);
	status->progress = gtk_progress_bar_new();
	gtk_widget_show(status->label);
	gtk_widget_show(status->progress);
	net_window_pack(NET_WINDOW(status->netwindow), status->progress);
	net_window_pack(NET_WINDOW(status->netwindow), status->label);

	status->js = journal_store_new_from_current(&err);

	if (!run_syncitems(status))
		goto out;

	if (!status->syncitems)
		goto out;

	st.lastsync = get_lastsync(status->js);

	if (!status->js) {
		net_window_show_error(NET_WINDOW(status->netwindow), err->message);
		g_error_free(err);
		goto out;
	}

	if (run_getevents(status)) {
		;
	}

	if (!journal_store_flush(status->js, &err)) {
		net_window_show_error(NET_WINDOW(status->netwindow), err->message);
		g_error_free(err);
		goto out;
	}

out:
	if (status->js)
		journal_store_free(status->js);
	gtk_widget_destroy(status->netwindow);
}

