/* logjam - a GTK client for LiveJournal.
 * Copyright (C) 2000-2003 Evan Martin <evan@livejournal.com>
 *
 * vim: tabstop=4 shiftwidth=4 noexpandtab :
 */

#include "config.h"

#include <gtk/gtk.h>
#include <string.h>

#include "conf.h"
#include "util.h"
#include "usejournal.h"

struct _UseJournal {
	GtkButton parent;
	GtkWidget *label;
};

static void usejournal_init(UseJournal *uj);

/* gtk stuff */
GType
usejournal_get_type(void) {
	static GType uj_type = 0;
	if (!uj_type) {
		static const GTypeInfo uj_info = {
			sizeof (GtkButtonClass),
			NULL,
			NULL,
			NULL,
			NULL,
			NULL,
			sizeof (UseJournal),
			0,
			(GInstanceInitFunc) usejournal_init,
		};
		uj_type = g_type_register_static(GTK_TYPE_BUTTON, "UseJournal",
				&uj_info, 0);
	}
	return uj_type;
}

void usejournal_revert(UseJournal *uj) {
	string_replace(&conf.usejournal, NULL);
	usejournal_update(uj);
}

void usejournal_update(UseJournal *uj) {
	char *text;
	const char *journalname;

	if (conf.usejournal)
		journalname = conf.usejournal;
	else if (conf_cur_user())
		journalname = conf_cur_user()->username;
	else
		journalname = _("(none)");

	text = g_strdup_printf(_("Using journal: %s"), journalname);
	gtk_label_set_text(GTK_LABEL(uj->label), text);
	g_free(text);
}

static void
usejournal_activate(GtkWidget *w, UseJournal *uj) {
	const gchar *user;

	if (!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)))
		return;

	user = gtk_label_get_text(GTK_LABEL(GTK_BIN(w)->child));

	if (g_ascii_strcasecmp(user, conf_cur_user()->username) == 0) {
		string_replace(&conf.usejournal, NULL); /* post as normal user. */
	} else {
		string_replace(&conf.usejournal, g_strdup(user));
	}
	usejournal_update(uj);
}

/* "having a shared prefix" is a pretty vague concept.
 * we try a few heuristics. */
static int
sharedprefix(char *a, char *b) {
	const int minmatch = 4;
	int i = 0;
	char *ulpos;

	/* does this journal have an underscore? go from there. */
	ulpos = strchr(b, '_');
	if (ulpos) {
		if (strncmp(a, b, ulpos-b) == 0)
			return ulpos - b;
	}

	/* or if the shared prefix is at least minmatch chars... */
	while (*a++ == *b++)
		i++;
	return (i > minmatch) ? i : 0;
}

static GtkWidget*
build_usejournal_menu(UseJournal *uj) {
	GtkWidget *menu, *curmenu, *item, *label;
	GSList *group = NULL;
	GList *l;
	char *journal;
	char *curmenuprefix = NULL;
	char prefix[30];

	curmenu = menu = gtk_menu_new(); 

	item = gtk_radio_menu_item_new_with_label(group, conf_cur_user()->username);
	group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(item));
	g_signal_connect(G_OBJECT(item), "activate",
			G_CALLBACK(usejournal_activate), uj);

	/* if they haven't specified a usejournal, default to their username. */
	if (conf.usejournal == NULL) 
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), TRUE);

	gtk_widget_show(item);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);

	for (l = conf_cur_user()->usejournals; l != NULL; l = l->next) {
		journal = (char*)l->data;

		if (curmenuprefix) {
			/* try to match this item to the prefix. */
			if (sharedprefix(curmenuprefix, journal)) {
				/* match. */
			} else {
				curmenu = menu;
				curmenuprefix = NULL;
			}
		} 
		if (!curmenuprefix && l->next) {
			/* try to see if this begins a new prefix. */
			char *nextjournal = (char*)l->next->data;
			int ofs;
			ofs = sharedprefix(journal, nextjournal);
			if (ofs) {
				/* make a new submenu for these shared-prefix journals. */
				char *markup;
				memcpy(prefix, journal, ofs);
				prefix[ofs] = 0;
				markup = g_strdup_printf("<i>%s</i>", prefix);

				item = gtk_menu_item_new();
				label = gtk_label_new(NULL);
				gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
				gtk_label_set_markup(GTK_LABEL(label), markup);
				gtk_container_add(GTK_CONTAINER(item), label);
				gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
				curmenu = gtk_menu_new();
				curmenuprefix = prefix;
				gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), curmenu);
				gtk_widget_show_all(item);
			}
		}
		item = gtk_radio_menu_item_new_with_label(group, journal);
		if (conf.usejournal && strcmp(conf.usejournal, journal) == 0)
			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), TRUE);
		g_signal_connect(G_OBJECT(item), "activate",
				G_CALLBACK(usejournal_activate), uj);
		gtk_widget_show(item);
		gtk_menu_shell_append(GTK_MENU_SHELL(curmenu), item);

		group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(item));
	}

	return menu;
}

static void
uj_pressed(UseJournal *uj, GdkEventButton *e) {
	GtkWidget *menu;

	menu = build_usejournal_menu(uj);
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
			e->button, e->time);
}

static void
usejournal_init(UseJournal *uj) {
	gtk_button_set_relief(GTK_BUTTON(uj), GTK_RELIEF_NONE);
	g_signal_connect(G_OBJECT(uj), "button-press-event",
			G_CALLBACK(uj_pressed), NULL);
	uj->label = gtk_label_new("text here");
	//gtk_misc_set_alignment(GTK_MISC(uj->label), 1.0, 0.5);
	gtk_widget_show(uj->label);
	gtk_container_add(GTK_CONTAINER(uj), uj->label);
}

GtkWidget*
usejournal_new(void) {
	UseJournal *uj = USEJOURNAL(g_object_new(usejournal_get_type(), NULL));

	return GTK_WIDGET(uj);
}

